#!/usr/bin/env python
"""This module describe nxtomomill applications which are available  through
the silx launcher.

Your environment should provide a command `nxtomomill`. You can reach help with
`tomwer --help`, and check the version with `nxtomomill --version`.
"""

import logging
import sys
from collections import namedtuple

from silx.utils.launcher import Launcher as _Launcher

import nxtomomill.version
from nxtomomill.utils.io import deprecated_warning

logging.basicConfig()


DeprecationWarning = namedtuple(
    "DeprecationWarning", ["since", "reason", "replacement"]
)


class Launcher(_Launcher):
    """
    Manage launch of module.

    Provides an API to describe available commands and feature to display help
    and execute the commands.
    """

    def __init__(
        self, prog=None, usage=None, description=None, epilog=None, version=None
    ):
        super().__init__(
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            version=version,
        )
        self._deprecations = {}
        "deprecations with prog names as key and deprecation info as values"

    def add_command(
        self,
        name=None,
        module_name=None,
        description=None,
        command=None,
        deprecated=False,
        deprecated_since_version=None,
        deprecated_reason=None,
        deprecated_replacement=None,
    ):
        super().add_command(
            name=name, module_name=module_name, description=description, command=command
        )
        if deprecated:
            self._deprecations[name] = DeprecationWarning(
                since=deprecated_since_version,
                reason=deprecated_reason,
                replacement=deprecated_replacement,
            )

    def execute(self, argv=None):
        if argv is None:
            argv = sys.argv

        if len(argv) <= 1:
            self.print_help()
            return 0

        command_name = argv[1]

        if command_name in self._deprecations:
            deprecation_info = self._deprecations[command_name]
            deprecated_warning(
                type_="application",
                name=command_name,
                reason=deprecation_info.reason,
                replacement=deprecation_info.replacement,
                since_version=deprecation_info.since,
            )
        super().execute(argv=argv)

    def print_help(self):
        """Print the help to stdout."""
        usage = self.usage
        if usage is None:
            usage = "usage: {0.prog} [--version|--help] <command> [<args>]"
        description = self.description
        epilog = self.epilog
        if epilog is None:
            epilog = "See '{0.prog} help <command>' to read about a specific subcommand"

        print(usage.format(self))
        print("")
        if description is not None:
            print(description)
            print("")
        print("The {0.prog} commands are:".format(self))
        commands = sorted(self._commands.keys())
        for command in commands:
            command = self._commands[command]
            print("   {:15} {:s}".format(command.name, command.description))
        print("")
        print(epilog.format(self))


def main():
    """Main function of the launcher

    This function is referenced in the pyproject.toml file, to create a
    launcher script generated by setuptools.

    :returns: The execution status
    """
    _version = nxtomomill.version.version
    launcher = Launcher(prog="nxtomomill", version=_version)
    launcher.add_command(
        "nx-copy",
        module_name="nxtomomill.app.nxcopy",
        description="Copy one or several NXtomo to another location",
    )
    launcher.add_command(
        "dxfile2nx",
        module_name="nxtomomill.app.dxfile2nx",
        description="Convert dx file to NXTomo",
    )
    launcher.add_command(
        "patch-nx",
        module_name="nxtomomill.app.patch_nx",
        description="allow to patch an NXTomo entry",
    )
    launcher.add_command(
        "tomoedf2nx",
        module_name="nxtomomill.app.edf2nx",
        description="convert spec-edf acquisition to nexus - hdf5",
        deprecated=True,
        deprecated_reason="remove `tomo` repetition. The shorter the better",
        deprecated_replacement="edf2nx",
        deprecated_since_version=0.5,
    )
    launcher.add_command(
        "edf2nx",
        module_name="nxtomomill.app.edf2nx",
        description="convert spec-edf acquisition to nexus - hdf5"
        "format to nx compliant file format",
    )
    launcher.add_command(
        "edf2nx-check",
        module_name="nxtomomill.app.edf2nx_check",
        description="check a conversion from EDF to Nxtomo and optionally remove EDF files",
    )
    launcher.add_command(
        "edf-quick-start",
        module_name="nxtomomill.app.edfconfig",
        description="create a configuration file to convert from spec-edf to NXtomo",
        deprecated_replacement="edf-config",
        deprecated_since_version=0.13,
    )
    launcher.add_command(
        "edf-config",
        module_name="nxtomomill.app.edfconfig",
        description="create a configuration file to convert from spec-edf to NXtomo",
    )
    launcher.add_command(
        "fluo-config",
        module_name="nxtomomill.app.fluoconfig",
        description="create a configuration file to convert from fluo-tomo to NXtomo",
    )
    launcher.add_command(
        "fluo2nx",
        module_name="nxtomomill.app.fluo2nx",
        description="Converts from XRFCT data to NXtomo",
    )
    launcher.add_command(
        "blissfluo-config",
        module_name="nxtomomill.app.blissfluoconfig",
        description="create a configuration file to convert from blissfluo-tomo to NXtomo",
    )
    launcher.add_command(
        "blissfluo2nx",
        module_name="nxtomomill.app.blissfluo2nx",
        description="Converts from (Bliss) XRFCT data to NXtomo",
    )
    launcher.add_command(
        "tomoh52nx",
        module_name="nxtomomill.app.h52nx",
        description="convert bliss hdf5 to nexus hdf5",
        deprecated=True,
        deprecated_reason="remove `tomo` repetition. The shorter the better",
        deprecated_replacement="h52nx",
        deprecated_since_version=0.5,
    )
    launcher.add_command(
        "h52nx",
        module_name="nxtomomill.app.h52nx",
        description="convert bliss hdf5 to nexus hdf5",
    )
    launcher.add_command(
        "z-concatenate-scans",
        module_name="nxtomomill.app.z_concatenate_scans",
        description="Build a concatenation nexus file from a z-series",
    )
    launcher.add_command(
        "zstages2nxs",
        module_name="nxtomomill.app.zstages2nxs",
        description="""creates all the target NXTomos for all the stages. Possibly reducing also reference scans for obtaining flats/dark""",
    )
    launcher.add_command(
        "h5-quick-start",
        module_name="nxtomomill.app.h5config",
        description="Create a default configuration file",
        deprecated_replacement="h5-config",
        deprecated_since_version=0.13,
    )
    launcher.add_command(
        "h5-config",
        module_name="nxtomomill.app.h5config",
        description="Create a default configuration file",
    )
    launcher.add_command(
        "split-nxfile",
        module_name="nxtomomill.app.split_nxfile",
        description="Split a file containing several NXtomo into a one file per NXtomo",
    )
    launcher.add_command(
        "fscan2nx",
        module_name="nxtomomill.app.fscan2nx",
        description="convert fscan hdf5 to nexus hdf5",
    )
    status = launcher.execute(sys.argv)
    return status


if __name__ == "__main__":
    # executed when using python -m PROJECT_NAME
    status = main()
    sys.exit(status)



# coding: utf-8

import os
import sys
import tempfile

import h5py
import numpy
import pint

from tomoscan.esrf.scan.utils import cwd_context

from nxtomomill import converter
from nxtomomill.tests.datasets import GitlabDataset
from nxtomomill.converter.hdf5.acquisition.utils import (
    get_nx_detectors,
    guess_nx_detector,
)
from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.tests.utils.bliss import _BlissSample
from nxtomomill.utils.hdf5 import DatasetReader, EntryReader
from nxtomomill.converter.hdf5.utils import (
    PROCESSED_DATA_DIR_NAME,
    RAW_DATA_DIR_NAME,
    get_default_output_file,
)
from nxtomomill.models.h52nx import H52nxModel

from tomoscan.io import HDF5File, get_swmr_mode

try:
    from tomoscan.esrf.scan.hdf5scan import NXtomoScan
except ImportError:
    from tomoscan.esrf.hdf5scan import NXtomoScan

import subprocess  # nosec B404
from glob import glob

import pytest
from silx.io.url import DataUrl
from silx.io.utils import get_data
from tomoscan.validator import is_valid_for_reconstruction

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.models.h52nx.FrameGroup import FrameGroup
from nxtomomill.tests.utils.bliss import MockBlissAcquisition

_ureg = pint.get_application_registry()


def url_has_been_copied(file_path: str, url: DataUrl):
    """util function to parse the `duplicate_data` folder and
    insure the copy of the dataset has been done"""
    duplicate_data_url = DataUrl(
        file_path=file_path, data_path="/duplicate_data", scheme="silx"
    )
    url_path = url.path()
    with EntryReader(duplicate_data_url) as duplicate_data_node:
        for _, dataset in duplicate_data_node.items():
            if "original_url" in dataset.attrs:
                original_url = dataset.attrs["original_url"]
                # the full dataset is registered in the attributes.
                # Here we only check the scan entry name
                if original_url.startswith(url_path):
                    return True
    return False


def test_simple_converter_with_nx_detector_attr(tmp_path):
    """
    Test a simple conversion when NX_class is defined
    """
    folder = tmp_path / "output_test_simple_converter_with_nx_detector_attr"
    folder.mkdir()
    config = TomoHDF5Config()
    config.no_master_file = False

    bliss_mock = MockBlissAcquisition(
        n_sample=2,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="pcolinux",
        create_tomo_config=False,
    )
    for sample in bliss_mock.samples:
        assert os.path.exists(sample.sample_file)
        config.output_file = sample.sample_file.replace(".h5", ".nx")
        config.input_file = sample.sample_file
        config.raises_error = True
        config.sample_x_keys = ("sx",)
        config.sample_y_keys = ("sy",)
        assert len(converter.get_bliss_tomo_entries(sample.sample_file, config)) == 1

        converter.from_h5_to_nx(
            configuration=config,
        )
        # insure only one file is generated
        assert os.path.exists(config.output_file)
        # insure data is here
        with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5s:
            for _, entry_node in h5s.items():
                assert "instrument/detector/data" in entry_node
                dataset = entry_node["instrument/detector/data"]
            # check virtual dataset are relative and valid
            assert dataset.is_virtual
            for vs in dataset.virtual_sources():
                assert not os.path.isabs(vs.file_name)
            # insure connection is valid. There is no
            # 'VirtualSource.is_valid' like function
            assert not (dataset[()].min() == 0 and dataset[()].max() == 0)
            instrument_grp = entry_node.require_group("instrument")
            assert "beam" in instrument_grp


def test_invalid_tomo_n(tmp_path):
    """Test translation fails if no detector can be found"""
    folder = tmp_path / "output_test_test_invalid_tomo_n"
    folder.mkdir()
    config = TomoHDF5Config()
    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        create_tomo_config=False,
    )
    assert len(bliss_mock.samples) == 1
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)
    output_file = sample.sample_file.replace(".h5", ".nx")

    # rewrite tomo_n
    with HDF5File(sample.sample_file, mode="a") as h5s:
        for _, entry_node in h5s.items():
            if "technique/scan/tomo_n" in entry_node:
                del entry_node["technique/scan/tomo_n"]
                entry_node["technique/scan/tomo_n"] = 9999

    with pytest.raises(ValueError):
        config.input_file = sample.sample_file
        config.output_file = output_file
        config.single_file = True
        config.no_input = True
        config.raises_error = True
        converter.from_h5_to_nx(configuration=config)


@pytest.mark.parametrize("create_tomo_config", (True, False))
def test_simple_converter_without_nx_detector_attr(tmp_path, create_tomo_config: bool):
    """
    Test a simple conversion when no NX_class is defined
    """
    folder = tmp_path / "output_test_simple_converter_without_nx_detector_attr"
    folder.mkdir()
    config = TomoHDF5Config()
    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=3,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="tata_detector",
        create_tomo_config=create_tomo_config,
    )
    assert len(bliss_mock.samples) == 1
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)
    output_file = sample.sample_file.replace(".h5", ".nx")
    config.input_file = sample.sample_file
    config.output_file = output_file
    config.single_file = True
    config.no_input = True
    converter.from_h5_to_nx(configuration=config)
    # insure only one file is generated
    assert os.path.exists(output_file)
    # insure data is here
    with HDF5File(output_file, mode="r", swmr=get_swmr_mode()) as h5s:
        for _, entry_node in h5s.items():
            assert "instrument/detector/data" in entry_node
            dataset = entry_node["instrument/detector/data"]
        # check virtual dataset are relative and valid
        assert dataset.is_virtual
        for vs in dataset.virtual_sources():
            assert not os.path.isabs(vs.file_name)
        # insure connection is valid. There is no
        # 'VirtualSource.is_valid' like function
        assert not (dataset[()].min() == 0 and dataset[()].max() == 0)
        # check NXdata group
        assert "data/data" in entry_node
        assert not (
            entry_node["data/data"][()].min() == 0
            and entry_node["data/data"][()].max() == 0
        )
        assert "data/rotation_angle" in entry_node
        assert "data/image_key" in entry_node


def test_providing_existing_camera_name(tmp_path):
    """Test that detector can be provided to the h5_to_nx function and
    using linux wildcard"""
    folder = tmp_path / "output_test_providing_existing_camera_name"
    folder.mkdir()
    config = TomoHDF5Config()

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=3,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="my_detector",
        create_tomo_config=False,
    )
    assert len(bliss_mock.samples) == 1
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)
    config.output_file = sample.sample_file.replace(".h5", ".nx")
    config.valid_camera_names = ("my_detec*",)
    config.input_file = sample.sample_file
    config.single_file = True
    config.no_input = True
    config.raises_error = True
    config.rotation_angle_keys = ("hrsrot",)
    config.sample_x_keys = ("sx",)
    config.sample_y_keys = ("sy",)
    converter.from_h5_to_nx(configuration=config)
    # insure only one file is generated
    assert os.path.exists(config.output_file)
    # insure data is here
    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5s:
        for _, entry_node in h5s.items():
            assert "instrument/detector/data" in entry_node
            dataset = entry_node["instrument/detector/data"]
        # check virtual dataset are relative and valid
        assert dataset.is_virtual
        for vs in dataset.virtual_sources():
            assert not os.path.isabs(vs.file_name)
        # insure connection is valid. There is no
        # 'VirtualSource.is_valid' like function
        assert not (dataset[()].min() == 0 and dataset[()].max() == 0)


def test_providing_non_existing_camera_name_no_tomo_config(tmp_path):
    """Test translation fails if no detector can be found"""
    folder = tmp_path / "output_test_providing_non_existing_camera_name_no_tomo_config"
    folder.mkdir()
    config = TomoHDF5Config()

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=3,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="toto_detector",
        create_tomo_config=False,
        z_series_v_3_options={
            "dark_at_start": True,
            "flat_at_start": True,
            "dark_at_end": False,
            "flat_at_end": False,
        },
    )
    assert len(bliss_mock.samples) == 1
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)
    config.input_file = sample.sample_file
    config.output_file = sample.sample_file.replace(".h5", ".nx")
    config.valid_camera_names = ("my_detec",)
    config.raises_error = True
    with pytest.raises(ValueError):
        converter.from_h5_to_nx(configuration=config)


@pytest.mark.parametrize("z_series_version", ("z-series-v1", "z-series-v3"))
def test_z_series_conversion_no_tomo_config(tmp_path, z_series_version: str):
    """Test conversion of a zseries bliss (mock) acquisition"""
    folder = tmp_path / "output_test_z_series_conversion_no_tomo_config"
    folder.mkdir()
    config = TomoHDF5Config()

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="frelon1",
        acqui_type=z_series_version,
        z_values=(1, 2, 3),
        create_tomo_config=False,
        z_series_v_3_options={
            "dark_at_start": True,
            "flat_at_start": True,
            "dark_at_end": False,
            "flat_at_end": False,
        },
    )
    assert len(bliss_mock.samples) == 1
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)
    config.input_file = sample.sample_file
    config.output_file = sample.sample_file.replace(".h5", ".nx")
    res = converter.from_h5_to_nx(configuration=config)
    # insure the 3 files are generated: one per z
    files = glob(os.path.dirname(sample.sample_file) + "/*.nx")
    assert len(files) == 3
    # try to create NXtomoScan from those to insure this is valid
    # and check z values for example
    for res_tuple in res:
        scan = NXtomoScan(scan=res_tuple[0], entry=res_tuple[1])
        if hasattr(scan, "translation_z"):
            assert scan.translation_z is not None
        assert is_valid_for_reconstruction(scan)


@pytest.mark.parametrize("z_series_version", ("z-series-v1", "z-series-v3"))
def test_z_series_conversion(tmp_path, z_series_version):
    """Test conversion of a zseries bliss (mock) acquisition"""
    folder = tmp_path / "output_test_z_series_conversion"
    folder.mkdir()
    config = TomoHDF5Config()

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="frelon1",
        acqui_type=z_series_version,
        z_values=(1, 2, 3),
        create_tomo_config=True,
        ebs_tomo_version="2.1.0",
        z_series_v_3_options={
            "dark_at_start": True,
            "flat_at_start": True,
            "dark_at_end": False,
            "flat_at_end": False,
        },
    )
    assert len(bliss_mock.samples) == 1
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)
    config.input_file = sample.sample_file
    config.output_file = sample.sample_file.replace(".h5", ".nx")
    res = converter.from_h5_to_nx(configuration=config)
    # insure the 3 files are generated: one per z
    files = glob(os.path.dirname(sample.sample_file) + "/*.nx")
    assert len(files) == 3
    # try to create NXtomoScan from those to insure this is valid
    # and check z values for example
    for res_tuple in res:
        scan = NXtomoScan(scan=res_tuple[0], entry=res_tuple[1])
        if hasattr(scan, "translation_z"):
            assert scan.translation_z is not None
        assert is_valid_for_reconstruction(scan)


def test_ignore_sub_entries(tmp_path):
    """
    Test we can ignore some sub entries
    """
    folder = tmp_path / "output_test_ignore_sub_entries"
    folder.mkdir()
    config = TomoHDF5Config()

    from nxtomomill.tests.utils.bliss import MockBlissAcquisition

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=0,
        n_flats=0,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="pcolinux",
    )
    for sample in bliss_mock.samples:
        assert os.path.exists(sample.sample_file)
        config.output_file = sample.sample_file.replace(".h5", ".nx")
        config.input_file = sample.sample_file
        config.single_file = True
        config.sub_entries_to_ignore = ("6.1", "7.1")
        config.no_input = True
        converter.from_h5_to_nx(configuration=config)
        # insure only one file is generated
        assert os.path.exists(config.output_file)
        # insure data is here
        with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5s:
            for _, entry_node in h5s.items():
                assert "instrument/detector/data" in entry_node
                dataset = entry_node["instrument/detector/data"]
            # check virtual dataset are relative and valid
            assert dataset.is_virtual
            assert dataset.shape == (
                10 * (10 - len(config.sub_entries_to_ignore)),
                64,
                64,
            )
            for vs in dataset.virtual_sources():
                assert not os.path.isabs(vs.file_name)
            # insure connection is valid. There is no
            # 'VirtualSource.is_valid' like function
            assert not (dataset[()].min() == 0 and dataset[()].max() == 0)


def create_nx_detector(node: h5py.Group, name, with_nx_class):
    det_node = node.require_group(name)
    data = numpy.random.random(10 * 10 * 10).reshape(10, 10, 10)
    det_node["data"] = data
    if with_nx_class:
        if "NX_class" not in det_node.attrs:
            det_node.attrs["NX_class"] = "NXdetector"
    return


def test_get_nx_detectors(tmp_path):
    """test get_nx_detectors function"""
    folder = tmp_path / "output_test_get_nx_detectors"
    folder.mkdir()

    h5file = os.path.join(folder, "h5file.hdf5")
    with HDF5File(h5file, mode="w") as h5s:
        create_nx_detector(node=h5s, name="det1", with_nx_class=True)
        create_nx_detector(node=h5s, name="det2", with_nx_class=False)
    with HDF5File(h5file, mode="r", swmr=get_swmr_mode()) as h5s:
        dets = get_nx_detectors(h5s)
        assert len(dets) == 1
        assert dets[0].name == "/det1"
        assert len(guess_nx_detector(h5s)) == 2
    with HDF5File(h5file, mode="a") as h5s:
        create_nx_detector(node=h5s, name="det3", with_nx_class=True)
        create_nx_detector(node=h5s, name="det4", with_nx_class=True)
    with HDF5File(h5file, mode="r", swmr=get_swmr_mode()) as h5s:
        dets = get_nx_detectors(h5s)
        assert len(dets) == 3


def test_guess_nx_detector(tmp_path):
    """test guess_nx_detector function"""
    folder = tmp_path / "output_test_guess_nx_detector"
    folder.mkdir()

    h5file = os.path.join(folder, "h5file.hdf5")

    with HDF5File(h5file, mode="w") as h5s:
        create_nx_detector(node=h5s, name="det2", with_nx_class=False)
    with HDF5File(h5file, mode="r", swmr=get_swmr_mode()) as h5s:
        dets = get_nx_detectors(h5s)
        assert len(dets) == 0
        dets = guess_nx_detector(h5s)
        assert dets[0].name == "/det2"
    with HDF5File(h5file, mode="w") as h5s:
        create_nx_detector(node=h5s, name="det3", with_nx_class=False)
        create_nx_detector(node=h5s, name="det4", with_nx_class=True)
    with HDF5File(h5file, mode="a") as h5s:
        dets = guess_nx_detector(h5s)
        assert len(dets) == 2


def create_scan(n_projection_scans, n_flats, n_darks, output_dir, frame_data_type):
    """
    :param int n_projection_scans: number of scans beeing projections
    :param int n_flats: number of frame per flats
    :param int n_darks: number of frame per dark
    """
    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=n_projection_scans,
        n_darks=n_darks,
        n_flats=n_flats,
        with_nx_detector_attr=True,
        output_dir=output_dir,
        detector_name="pcolinux",
        frame_data_type=frame_data_type,
    )
    return bliss_mock.samples[0].sample_file


def test_dataset_1(tmp_path):
    """test a conversion where projections are contained in the
    input_file. Dark and flats are on a different file"""

    frame_data_type = numpy.uint16
    folder = tmp_path / "output_test_dataset_1"
    folder.mkdir()
    config = TomoHDF5Config()
    config.no_master_file = False
    config.output_file = os.path.join(folder, "output.nx")
    config.rotation_angle_keys = ("hrsrot",)
    config.sample_x_keys = ("sx",)
    config.sample_y_keys = ("sy",)

    folder_1 = os.path.join(folder, "acqui_1")
    input_file = create_scan(
        n_projection_scans=6,
        n_flats=0,
        n_darks=0,
        output_dir=folder_1,
        frame_data_type=frame_data_type,
    )
    folder_2 = os.path.join(folder, "acqui_2")
    dark_flat_file = create_scan(
        n_projection_scans=0,
        n_flats=1,
        n_darks=1,
        output_dir=folder_2,
        frame_data_type=frame_data_type,
    )
    config.input_file = input_file

    # we want to take two scan of projections from the input file: 5.1
    # and 6.1. As the input file is provided we don't need to
    # specify it
    config.data_scans = (
        FrameGroup(frame_type="proj", url=DataUrl(data_path="/5.1", scheme="silx")),
        FrameGroup(frame_type="proj", url=DataUrl(data_path="/6.1", scheme="silx")),
        FrameGroup(
            frame_type="flat",
            url=DataUrl(file_path=dark_flat_file, data_path="/2.1", scheme="silx"),
        ),
        FrameGroup(
            frame_type="dark",
            url=DataUrl(file_path=dark_flat_file, data_path="/3.1", scheme="silx"),
        ),
    )
    converter.from_h5_to_nx(
        configuration=config,
    )

    assert os.path.exists(config.output_file), "output file does not exists"

    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5s:
        assert len(h5s.items()) == 1
        assert "entry0000" in h5s

    scan = NXtomoScan(scan=config.output_file, entry="entry0000")
    assert is_valid_for_reconstruction(scan)

    # check the `data`has been created
    assert len(scan.projections) == 20
    assert len(scan.darks) == 10
    # check data is a virtual dataset
    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5f:
        dataset = h5f["entry0000/instrument/detector/data"]
        assert dataset.is_virtual
        assert dataset.dtype == frame_data_type
    # check the `data` virtual dataset is valid
    # if the link fail then all values are zeros
    url = tuple(scan.projections.values())[0]
    proj_data = get_data(url)
    assert proj_data.min() != proj_data.max()

    url = tuple(scan.darks.values())[0]
    dark_data = get_data(url)
    assert dark_data.min() != dark_data.max()

    assert len(scan.flats) == 10
    url = tuple(scan.flats.values())[0]
    flat_data = get_data(url)
    assert flat_data.min() != flat_data.max()


def test_dataset_2(tmp_path):
    """test a conversion where no input file is provided and
    where we have 2 projections in a file, 3 in an other.
    Flat and darks are also in another file. No flat provided.
    """
    folder = tmp_path / "output_test_dataset_2"
    folder.mkdir()
    frame_data_type = numpy.uint16
    config = TomoHDF5Config()
    config.no_master_file = False
    config.output_file = os.path.join(folder, "output.nx")
    config.rotation_angle_keys = ("hrsrot",)
    config.sample_x_keys = ("sx",)
    config.sample_y_keys = ("sy",)

    folder_1 = os.path.join(folder, "acqui_1")
    file_1 = create_scan(
        n_projection_scans=6,
        n_flats=0,
        n_darks=0,
        output_dir=folder_1,
        frame_data_type=frame_data_type,
    )
    folder_2 = os.path.join(folder, "acqui_2")
    file_2 = create_scan(
        n_projection_scans=6,
        n_flats=0,
        n_darks=0,
        output_dir=folder_2,
        frame_data_type=frame_data_type,
    )
    folder_3 = os.path.join(folder, "acqui_3")
    file_3 = create_scan(
        n_projection_scans=0,
        n_flats=0,
        n_darks=1,
        output_dir=folder_3,
        frame_data_type=frame_data_type,
    )
    folder_4 = os.path.join(folder, "acqui_4")
    file_4 = create_scan(
        n_projection_scans=0,
        n_flats=1,
        n_darks=0,
        output_dir=folder_4,
        frame_data_type=frame_data_type,
    )

    # we want to take two scan of projections from the input file: 5.1
    # and 6.1. As the input file is provided we don't need to
    # specify it
    dark_url_1 = DataUrl(file_path=file_3, data_path="/2.1", scheme="silx")
    flat_url_1 = DataUrl(file_path=file_4, data_path="/2.1", scheme="silx")
    proj_url_1 = DataUrl(file_path=file_1, data_path="/5.1", scheme="silx")
    proj_url_2 = DataUrl(file_path=file_1, data_path="/6.1", scheme="silx")
    proj_url_3 = DataUrl(file_path=file_2, data_path="/4.1", scheme="silx")
    proj_url_4 = DataUrl(file_path=file_2, data_path="/2.1", scheme="silx")

    config.default_data_copy = True
    config.data_scans = (
        FrameGroup(frame_type="dark", url=dark_url_1),
        FrameGroup(frame_type="flat", url=flat_url_1),
        FrameGroup(frame_type="proj", url=proj_url_1, copy_data=False),
        FrameGroup(frame_type="proj", url=proj_url_2, copy_data=False),
        FrameGroup(frame_type="proj", url=proj_url_3),
        FrameGroup(frame_type="proj", url=proj_url_4),
    )

    urls_copied = (dark_url_1, flat_url_1, proj_url_3, proj_url_4)
    urls_not_copied = (proj_url_1, proj_url_2)

    config.raises_error = True
    converter.from_h5_to_nx(
        configuration=config,
    )

    assert os.path.exists(config.output_file)

    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5s:
        assert "entry0000" in h5s
        assert len(h5s.items()) == 1

    with HDF5File(
        config.output_file.replace(".nx", "_0000.nx"),
        mode="r",
        swmr=get_swmr_mode(),
    ) as h5s:
        assert "entry0000" in h5s
        assert "duplicate_data" in h5s
        assert len(h5s.items()) == 2

    detector_url = DataUrl(
        file_path=config.output_file,
        data_path="/entry0000/instrument/detector/data",
        scheme="silx",
    )
    with DatasetReader(detector_url) as detector_dataset:
        assert detector_dataset.is_virtual
        for i_vs, vs in enumerate(detector_dataset.virtual_sources()):
            assert not os.path.isabs(vs.file_name)
            if i_vs in (0, 1, 4, 5):
                assert vs.file_name == "."
            else:
                assert vs.file_name == "./acqui_1/sample_0/sample_0.h5"
    # FIXME: avoid keeping some file open. not clear why this is needed
    detector_dataset = None

    scan = NXtomoScan(scan=config.output_file, entry="entry0000")
    assert is_valid_for_reconstruction(scan)

    # check the `data`has been created
    assert len(scan.projections) == 40
    assert len(scan.darks) == 10

    # check the `data` virtual dataset is valid
    # if the link fail then all values are zeros
    url = tuple(scan.projections.values())[0]
    proj_data = get_data(url)
    assert proj_data.min() != proj_data.max()

    url = tuple(scan.darks.values())[0]
    dark_data = get_data(url)
    assert dark_data.min() != dark_data.max()

    assert len(scan.flats) == 10
    url = tuple(scan.flats.values())[0]
    flat_data = get_data(url)
    assert flat_data.min() != flat_data.max()

    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5f:
        dataset = h5f["entry0000/instrument/detector/data"][()]
        assert dataset.shape[0] == 60
    with EntryReader(dark_url_1) as dark_entry:
        numpy.testing.assert_array_equal(
            dark_entry["instrument/pcolinux/data"], dataset[0:10]
        )
    with EntryReader(flat_url_1) as flat_entry:
        numpy.testing.assert_array_equal(
            flat_entry["instrument/pcolinux/data"], dataset[10:20]
        )
    with EntryReader(proj_url_1) as proj_entry_1:
        numpy.testing.assert_array_equal(
            proj_entry_1["instrument/pcolinux/data"], dataset[20:30]
        )
    with EntryReader(proj_url_2) as proj_entry_2:
        numpy.testing.assert_array_equal(
            proj_entry_2["instrument/pcolinux/data"], dataset[30:40]
        )
    with EntryReader(proj_url_3) as proj_entry_3:
        numpy.testing.assert_array_equal(
            proj_entry_3["instrument/pcolinux/data"], dataset[40:50]
        )
    with EntryReader(proj_url_4) as proj_entry_4:
        numpy.testing.assert_array_equal(
            proj_entry_4["instrument/pcolinux/data"], dataset[50:60]
        )

    for url in urls_copied:
        assert url_has_been_copied(
            file_path=config.output_file.replace(".nx", "_0000.nx"),
            url=url,
        )

    for url in urls_not_copied:
        assert not url_has_been_copied(
            file_path=config.output_file.replace(".nx", "_0000.nx"),
            url=url,
        )

    # test with some extra parameters
    config.energy_kev = 12.2
    config.x_sample_pixel_size_m = 2.6 * 10e-6
    config.y_sample_pixel_size_m = 2.7 * 10e-6

    config.overwrite = True
    config.field_of_view = "Half"

    init_url_1 = DataUrl(file_path=file_1, data_path="/1.1", scheme="silx")

    config.default_data_copy = True
    config.data_scans = (
        FrameGroup(frame_type="init", url=init_url_1),
        FrameGroup(frame_type="dark", url=dark_url_1),
        FrameGroup(frame_type="flat", url=flat_url_1),
        FrameGroup(frame_type="proj", url=proj_url_1, copy=False),
        FrameGroup(frame_type="proj", url=proj_url_2, copy=False),
        FrameGroup(frame_type="proj", url=proj_url_3),
        FrameGroup(frame_type="proj", url=proj_url_4),
    )

    converter.from_h5_to_nx(
        configuration=config,
    )
    scan.clear_cache()
    energy = scan.energy
    assert numpy.isclose(energy, 12.2)
    assert scan.sample_x_pixel_size is not None
    assert numpy.isclose(scan.sample_x_pixel_size, 2.6 * 10e-6)
    assert scan.sample_y_pixel_size is not None
    assert numpy.isclose(scan.sample_y_pixel_size, 2.7 * 10e-6)
    with EntryReader(
        DataUrl(file_path=scan.master_file, data_path=scan.entry, scheme="h5py")
    ) as entry:
        assert "instrument/detector" in entry
        assert "instrument/diode" not in entry


@pytest.mark.parametrize("z_series_version", ("z-series-v1",))
def test_z_series_conversion_with_external_urls(tmp_path, z_series_version: str):
    """
    test conversion of a z-series using configuration
    """

    folder = tmp_path / "test_z_series_conversion_with_external_urls"
    folder = tempfile.mkdtemp()
    frame_data_type = numpy.uint64

    config = TomoHDF5Config()
    config.output_file = os.path.join(folder, "output.nx")

    # dataset init
    camera_name = "frelon"
    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=2,
        n_darks=1,
        n_flats=0,
        with_nx_detector_attr=True,
        output_dir=os.path.join(folder, "seq_1"),
        detector_name=camera_name,
        acqui_type=z_series_version,
        z_values=(1, 2, 3),
        frame_data_type=frame_data_type,
    )
    zseries_1_file = bliss_mock.samples[0].sample_file

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=2,
        n_darks=0,
        n_flats=1,
        with_nx_detector_attr=True,
        output_dir=os.path.join(folder, "seq_2"),
        detector_name=camera_name,
        acqui_type=z_series_version,
        z_values=(4, 5, 6),
        frame_data_type=frame_data_type,
    )
    zseries_2_file = bliss_mock.samples[0].sample_file

    dark_url_1 = DataUrl(file_path=zseries_1_file, data_path="/5.1", scheme="silx")
    proj_url_1 = DataUrl(file_path=zseries_1_file, data_path="/6.1", scheme="silx")
    proj_url_2 = DataUrl(file_path=zseries_1_file, data_path="/7.1", scheme="silx")
    proj_url_3 = DataUrl(file_path=zseries_1_file, data_path="/9.1", scheme="silx")
    proj_url_4 = DataUrl(file_path=zseries_1_file, data_path="/10.1", scheme="silx")
    proj_url_5 = DataUrl(file_path=zseries_2_file, data_path="/3.1", scheme="silx")
    proj_url_6 = DataUrl(file_path=zseries_2_file, data_path="/4.1", scheme="silx")
    flat_url_1 = DataUrl(file_path=zseries_2_file, data_path="/2.1", scheme="silx")
    config.default_data_copy = True
    config.single_file = True
    config.data_scans = (
        FrameGroup(frame_type="dark", url=dark_url_1, copy_data=False),
        FrameGroup(frame_type="flat", url=flat_url_1, copy_data=False),
        FrameGroup(frame_type="proj", url=proj_url_1),
        FrameGroup(frame_type="proj", url=proj_url_2),
        FrameGroup(frame_type="proj", url=proj_url_3),
        FrameGroup(frame_type="proj", url=proj_url_4),
        FrameGroup(frame_type="proj", url=proj_url_5),
        FrameGroup(frame_type="proj", url=proj_url_6),
    )
    urls_copied = (
        proj_url_1,
        proj_url_2,
        proj_url_3,
        proj_url_4,
        proj_url_5,
        proj_url_6,
    )
    urls_not_copied = (flat_url_1, dark_url_1)
    # do conversion
    new_scans = converter.from_h5_to_nx(
        configuration=config,
    )
    assert len(new_scans) == 3
    assert os.path.exists(config.output_file)
    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5f:
        assert "entry0000" in h5f
        assert "entry0001" in h5f
        assert "entry0002" in h5f

    scan_z0 = NXtomoScan(scan=config.output_file, entry="entry0000")
    scan_z1 = NXtomoScan(scan=config.output_file, entry="entry0001")
    scan_z2 = NXtomoScan(scan=config.output_file, entry="entry0002")

    # check the `data`has been created
    assert len(scan_z0.projections) == 20
    assert len(scan_z1.projections) == 20
    assert len(scan_z2.projections) == 20

    for url in urls_copied:
        assert url_has_been_copied(file_path=config.output_file, url=url)

    for url in urls_not_copied:
        assert not url_has_been_copied(file_path=config.output_file, url=url)

    # test a few slices
    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5f:
        dataset = h5f["entry0000/instrument/detector/data"]
        assert dataset.shape[0] == 30

        with EntryReader(proj_url_1) as proj_entry_1:
            numpy.testing.assert_array_equal(
                proj_entry_1["instrument/frelon/data"], dataset[10:20]
            )
        with EntryReader(proj_url_2) as proj_entry_2:
            numpy.testing.assert_array_equal(
                proj_entry_2["instrument/frelon/data"], dataset[20:30]
            )
        assert dataset.dtype == frame_data_type


@pytest.mark.parametrize("z_series_version", ("z-series-v1", "z-series-v3"))
@pytest.mark.parametrize(
    "dark_flat_config",
    (
        {
            "dark_at_start": True,
            "flat_at_start": True,
            "dark_at_end": False,
            "flat_at_end": False,
        },
        {
            "dark_at_start": False,
            "flat_at_start": False,
            "dark_at_end": True,
            "flat_at_end": True,
        },
    ),
)
def test_z_series_dark_flat_copy(tmp_path, z_series_version: str, dark_flat_config):
    """In z-series version 3"""

    folder = tmp_path / "h52nx_from_command_line"
    folder.mkdir()

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=2,
        n_darks=1,
        n_flats=1,
        acqui_type=z_series_version,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="pcolinux",
        frame_data_type=numpy.uint32,
        z_values=(0.0, 1.0, 2.0),
        z_series_v_3_options=dark_flat_config,
    )

    # launch conversion
    sample = bliss_mock.samples[0]
    input_file = sample.sample_file
    assert os.path.exists(input_file)

    config = TomoHDF5Config()
    config.output_file = get_default_output_file(sample.sample_file)
    config.valid_camera_names = ("pcolinux",)
    config.input_file = sample.sample_file
    config.single_file = True
    config.no_input = True
    config.raises_error = True
    config.rotation_angle_keys = ("hrsrot",)

    new_entries = converter.from_h5_to_nx(configuration=config)
    # insure only one file is generated
    assert os.path.exists(config.output_file)
    # insure data is here
    for i_nx_tomo, (file_path, data_path) in enumerate(new_entries):
        nx_tomo = NXtomo().load(file_path=file_path, data_path=data_path)
        assert (
            len(
                numpy.where(
                    nx_tomo.instrument.detector.image_key_control == ImageKey.DARK_FIELD
                )[0]
            )
            == 10
        ), f"NXtomo {i_nx_tomo} doesn't have the expected number of dark"
        assert (
            len(
                numpy.where(
                    nx_tomo.instrument.detector.image_key_control == ImageKey.FLAT_FIELD
                )[0]
            )
            == 10
        ), f"NXtomo {i_nx_tomo} doesn't have the expected number of flat"
        assert (
            len(
                numpy.where(
                    nx_tomo.instrument.detector.image_key_control == ImageKey.PROJECTION
                )[0]
            )
            == 20
        ), f"NXtomo {i_nx_tomo} doesn't have the expected number of projection"


def test_h52nx_from_command_line(tmp_path):

    folder = tmp_path / "h52nx_from_command_line"
    folder.mkdir()

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=2,
        n_darks=1,
        n_flats=1,
        with_nx_detector_attr=True,
        output_dir=folder,
        detector_name="pcolinux",
        frame_data_type=numpy.uint32,
    )
    sample = bliss_mock.samples[0]
    output_file = sample.sample_file.replace(".h5", ".nx")
    assert os.path.exists(sample.sample_file)

    with cwd_context(os.path.dirname(sample.sample_file)):

        input_file = os.path.basename(sample.sample_file)
        assert os.path.exists(input_file)
        output_file = os.path.basename(output_file)
        assert not os.path.exists(output_file)
        cmd = (
            sys.executable,
            "-m",
            "nxtomomill",
            "h52nx",
            input_file,
            output_file,
            "--copy-data",
            "--raises-error",
            "--single-file",
        )
        subprocess.call(cmd, cwd=os.path.dirname(sample.sample_file))  # nosec B603

        assert os.path.exists(output_file)

        # insure all link are connected to one file: the internal one
        frame_dataset_url = DataUrl(
            file_path=output_file,
            data_path="/entry0000/instrument/detector/data",
            scheme="silx",
        )
        with DatasetReader(frame_dataset_url) as dataset:
            assert dataset.is_virtual

            for vs_info in dataset.virtual_sources():
                assert dataset.is_virtual
                assert vs_info.file_name == "."
                assert dataset.dtype == numpy.uint32
        # FIXME: avoid keeping some file open. not clear why this is needed
        dataset = None
        with HDF5File(output_file, "r", swmr=get_swmr_mode()) as h5f:
            assert "/entry0000/instrument/diode" not in h5f

        # insure an nxtomo can be created from it
        nx_tomo = NXtomo().load(output_file, "entry0000")
        assert nx_tomo.energy is not None
        z1 = -52.0 * _ureg.meter
        z2 = 0.1 * _ureg.meter
        assert nx_tomo.instrument.detector.distance == z2
        assert nx_tomo.instrument.source.distance == z1
        propa_dist_read = nx_tomo.sample.propagation_distance.to_base_units().magnitude
        propa_dist_expected = ((-z1 * z2) / (-z1 + z2)).to_base_units().magnitude
        assert numpy.isclose(
            propa_dist_read,
            propa_dist_expected,
        )
        numpy.testing.assert_array_equal(
            nx_tomo.instrument.detector.sequence_number,
            numpy.linspace(0, 40, 40, endpoint=False, dtype=numpy.uint32),
        )


input_types = (
    numpy.uint8,
    numpy.uint16,
    numpy.uint32,
    numpy.uint64,
    numpy.float16,
    numpy.float32,
    numpy.int16,
    numpy.int32,
    numpy.int64,
)


@pytest.mark.parametrize("rotation_is_clockwise", (True, False))
@pytest.mark.parametrize("input_type", input_types)
def test_simple_conversion(input_type, tmp_path, rotation_is_clockwise: bool):
    """test simple conversion from different frame data type and handling of RAW_DATA with PROCESSED_DATA"""

    input_path = tmp_path / "test" / RAW_DATA_DIR_NAME / "dataset"
    os.makedirs(input_path)

    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=10,
        n_darks=5,
        n_flats=5,
        with_nx_detector_attr=True,
        output_dir=input_path,
        detector_name="my_detector",
        frame_data_type=input_type,
        rotation_is_clockwise=rotation_is_clockwise,
        create_tomo_config=True,
    )
    sample = bliss_mock.samples[0]
    assert os.path.exists(sample.sample_file)

    config = TomoHDF5Config()
    config.output_file = get_default_output_file(sample.sample_file)
    assert PROCESSED_DATA_DIR_NAME in config.output_file
    config.valid_camera_names = ("my_detec*",)
    config.input_file = sample.sample_file
    config.single_file = True
    config.no_input = True
    config.raises_error = True
    config.rotation_angle_keys = ("hrsrot",)

    converter.from_h5_to_nx(configuration=config)
    # insure only one file is generated
    assert os.path.exists(config.output_file)
    # insure data is here
    with HDF5File(config.output_file, mode="r", swmr=get_swmr_mode()) as h5s:
        for _, entry_node in h5s.items():
            assert "instrument/detector/data" in entry_node
            dataset = entry_node["instrument/detector/data"]
            assert dataset.dtype == input_type
            assert "control" not in entry_node
            if rotation_is_clockwise:
                assert numpy.all(
                    entry_node["data/rotation_angle"][()] <= 0.0
                ), "tomoconfig 'rotation_is_clockwise' doesn't seems to be properly read"
            else:
                assert numpy.all(
                    entry_node["data/rotation_angle"][()] >= 0.0
                ), "tomoconfig 'rotation_is_clockwise' doesn't seems to be properly read"


def test_machine_current():
    """Test machine current is handle by the convertor"""
    with tempfile.TemporaryDirectory() as root_dir:
        bliss_mock = MockBlissAcquisition(
            n_sample=1,
            n_sequence=1,
            n_scan_per_sequence=2,
            n_darks=5,
            n_flats=5,
            with_nx_detector_attr=True,
            output_dir=root_dir,
            detector_name="my_detector",
        )
        sample = bliss_mock.samples[0]
        assert os.path.exists(sample.sample_file)
        # append current to the bliss file
        # from the example file I had it looks like this information can be saved at different location
        # and can be either a number (for dark and flat for example) or a list (for projections)
        with HDF5File(sample.sample_file, mode="a") as h5f:
            # overwrite start_time to made ordering work
            del h5f["1.1"]["start_time"]
            h5f["1.1"]["start_time"] = "2022-01-15T21:05:58.360095+02:00"

        with HDF5File(sample.sample_file, mode="a") as h5f:
            node_names = ("2.1", "3.1")
            machine_current = (602, 589)  # those are in ma
            start_times = (
                "2022-01-15T21:07:58.360095+02:00",
                "2022-01-15T21:07:59.360095+02:00",
            )
            for node_name, machine_current_ma, st in zip(
                node_names, machine_current, start_times
            ):
                h5f[f"{node_name}/instrument/machine/current"] = machine_current_ma
                h5f[f"{node_name}/instrument/machine/current"].attrs["units"] = str(
                    _ureg.milliampere
                )
                del h5f[node_name]["start_time"]
                h5f[node_name]["start_time"] = st

            assert "4.1" in h5f
            assert "5.1" in h5f
            assert (
                "6.1" not in h5f
            )  # this is because n_scan_per_sequence == 2 in MockBlissAcquisition

            # create some X.2 for machine current as this is done in Bliss
            for node_name in ("4.2", "5.2"):
                h5f.require_group(node_name)["title"] = _BlissSample.get_title(
                    "projection"
                )

                current_monitor_dataset = h5f.require_dataset(
                    f"{node_name}/measurement/current", shape=(5), dtype=numpy.float32
                )
                current_monitor_dataset[:] = numpy.linspace(
                    0.9, 0.96, 5, dtype=numpy.float32, endpoint=True
                )
                # add a nan to make sure this is properly handled
                current_monitor_dataset[1] = numpy.nan
                current_monitor_dataset.attrs["units"] = str(_ureg.ampere)

            # define start_time and end_time to insure conversion is correct
            # start_time and end_time is required for both:
            #    * from X.1 to create frame time stamp
            #    * from X.2 to get machine current time stamp
            start_times = (
                "2022-01-15T21:08:58.360095+02:00",
                "2022-01-15T21:10:58.360095+02:00",
            )
            end_times = (
                "2022-01-15T21:09:58.360095+02:00",
                "2022-01-15T21:11:58.360095+02:00",
            )
            tuple_node_names = (["4.1", "4.2"], ["5.1", "5.2"])

            for node_names, st, et in zip(tuple_node_names, start_times, end_times):
                for node_name in node_names:
                    if "start_time" in h5f[node_name]:
                        del h5f[node_name]["start_time"]
                    h5f[node_name]["start_time"] = st
                    h5f[node_name]["end_time"] = et
                    h5f[node_name].require_group("instrument")

        # convert the file
        config = TomoHDF5Config()
        config.output_file = sample.sample_file.replace(".h5", ".nx")
        config.single_file = True
        config.no_input = True
        config.raises_error = True
        with pytest.raises(ValueError):
            converter.from_h5_to_nx(configuration=config)
        config.input_file = sample.sample_file
        converter.from_h5_to_nx(configuration=config)
        # insure only one file is generated
        assert os.path.exists(config.output_file)
        # insure data is here
        nx_tomo = NXtomo().load(config.output_file, "entry0000")
        expected_results = numpy.concatenate(
            [
                [602 / 1000] * 10,
                [589 / 1000] * 10,
                numpy.linspace(0.9, 0.96, 10, dtype=numpy.float32),
                numpy.linspace(0.9, 0.96, 10, dtype=numpy.float32),
            ]
        )
        n_frames = (
            10 * 4
        )  # there is 10 frames per scan. One dark, one flat and two projections scans
        assert len(nx_tomo.control.data) == n_frames
        numpy.testing.assert_allclose(
            nx_tomo.control.data, expected_results, rtol=0.001
        )

        # test also from tomoscan
        scan = NXtomoScan(scan=config.output_file, entry="entry0000")
        # check getting the projections machine current
        assert (
            len(
                scan.machine_current[
                    scan.image_key_control == ImageKey.PROJECTION.value
                ]
            )
            == 20
        )
        assert (
            len(
                scan.machine_current[
                    scan.image_key_control == ImageKey.DARK_FIELD.value
                ]
            )
            == 10
        )
        assert (
            len(
                scan.machine_current[
                    scan.image_key_control == ImageKey.FLAT_FIELD.value
                ]
            )
            == 10
        )


files_to_expected_result: dict[str, tuple[str]] = {
    "h5_datasets/multitomo/id19/20250713/034_oak_1__0001.h5": (
        [f"034_oak_1__0001_{str(i).zfill(4)}.nx" for i in range(140)]
    ),
    "h5_datasets/holotomo/id16b/20250711/B3P7_sponge_pco2_ht_25nm.h5": (
        "B3P7_sponge_pco2_ht_25nm_0000.nx",
        "B3P7_sponge_pco2_ht_25nm_0001.nx",
        "B3P7_sponge_pco2_ht_25nm_0002.nx",
        "B3P7_sponge_pco2_ht_25nm_0003.nx",
    ),
    "h5_datasets/holotomo/id16b/20250325/needle_test_alu_17p4_daiquiri_50nm_ht_0004.h5": (
        "needle_test_alu_17p4_daiquiri_50nm_ht_0004_0000.nx",
        "needle_test_alu_17p4_daiquiri_50nm_ht_0004_0001.nx",
        "needle_test_alu_17p4_daiquiri_50nm_ht_0004_0002.nx",
        "needle_test_alu_17p4_daiquiri_50nm_ht_0004_0003.nx",
    ),
}


@pytest.mark.parametrize(
    "input_file,expected_result_file_names", files_to_expected_result.items()
)
def test_output_file_indexing(
    input_file: str, expected_result_file_names: tuple[tuple[str, str]], tmp_path
):
    """test nxtomo indexing policy"""
    scan_dir = GitlabDataset.get_dataset(os.path.dirname(input_file))
    bliss_file = os.path.join(scan_dir, os.path.basename(input_file))

    output_file = (
        tmp_path / "output" / os.path.basename(input_file).replace(".h5", ".nx")
    )
    config = TomoHDF5Config()
    config.input_file = bliss_file
    config.output_file = str(output_file)

    converter.from_h5_to_nx(
        configuration=config,
    )
    assert tuple(sorted(os.listdir(tmp_path / "output"))) == tuple(
        sorted(expected_result_file_names)
    )


def test_id16a_use_case(tmp_path):
    """
    Dummy test to match the ID16a use case: providing the input and output file.
    Check that the rotation_is_clockwise works. Check the sample and detector pixel size.
    """
    output_file = tmp_path / "output.nx"
    config = H52nxModel()
    config.single_file = True
    # add noise to the input and output file to make sure they are overwritten
    config.input_file = "toto"
    config.output_file = "tata"

    input_file = GitlabDataset.get_dataset(
        "h5_datasets/holotomo/id16a/Atomium_S2_holo4_HT_010nm_0001.h5"
    )

    config_file = tmp_path / "config.cfg"
    config.to_cfg_file(file_path=str(config_file))

    cmd = (
        sys.executable,
        "-m",
        "nxtomomill",
        "h52nx",
        input_file,
        output_file,
        "--config",
        config_file,
    )
    subprocess.call(cmd, cwd=tmp_path)  # nosec B603

    assert os.path.exists(output_file)

    # check 4 NXtomo exists
    nxtomos = [
        NXtomo().load(
            file_path=output_file,
            data_path=f"entry000{i}",
        )
        for i in range(4)
    ]
    assert len(nxtomos) == 4

    nxtomo_d0 = nxtomos[0]
    assert len(nxtomo_d0.sample.rotation_angle) == 2064

    # check rotation_is_clockwise
    raw_rot_angles = numpy.concatenate(
        [
            [0.0] * 20,  # dark
            [0.0] * 20,  # flat
            numpy.linspace(0.0, 180.0, 2001, endpoint=True),  # proj
            [180.0, 90.0, 0.0],  # alignment proj
            [0.0] * 20,  # flat
        ]
    )

    numpy.testing.assert_almost_equal(
        nxtomo_d0.sample.rotation_angle.to("degree").magnitude,
        raw_rot_angles,
        decimal=2,
    )

    # check detector pixel size
    numpy.testing.assert_almost_equal(
        nxtomo_d0.instrument.detector.x_pixel_size.to(_ureg.micrometer).magnitude,
        2.95203,
        decimal=4,
    )
    numpy.testing.assert_almost_equal(
        nxtomo_d0.instrument.detector.y_pixel_size.to(_ureg.micrometer).magnitude,
        2.95203,
        decimal=4,
    )

    # check sample pixel size
    numpy.testing.assert_almost_equal(
        nxtomo_d0.sample.x_pixel_size.to(_ureg.micrometer).magnitude,
        0.00999999,
        decimal=4,
    )
    numpy.testing.assert_almost_equal(
        nxtomo_d0.sample.y_pixel_size.to(_ureg.micrometer).magnitude,
        0.00999999,
        decimal=4,
    )

    # check detector flips
    transformations = {
        t.axis_name: t
        for t in nxtomo_d0.instrument.detector.transformations.transformations
    }
    for transformation_name in ("rx", "ry"):
        transformation = transformations[transformation_name]
        transformation.transformation_values.magnitude == numpy.array([0.0])



# coding: utf-8

from __future__ import annotations

import datetime
import os

import numpy
import h5py
from tomoscan.io import HDF5File


class MockBlissAcquisition:
    """

    :param n_sequence: number of sequence to create
    :param n_scan_per_sequence: number of scans (projection series) per sequence
    :param n_projections_per_scan: number of projection frame in a scan
    :param n_darks: number of dark frame in the series. Only one series at the
                    beginning
    :param n_flats: number of flats to create. In this case will only
                        create one series of n flats after dark if any
    :param output_dir: will contain the proposal file and one folder per
                           sequence.
    :param str acqui_type: acquisition type. Can be "basic", "multitomo", "zseries-v1", "zseries-v3
    :param sequence z_values: if acqui_type is zseries then users should
                              provide the serie of values for z (one per stage)
    :param nb_loop: number of multitomo loop for v1 of bliss multitomo
    :param nb_tomo: number of tomo per loop for v1 of bliss multitomo
    :param nb_turns: number of turns for v2 of bliss multitomo ( <=> nb NXtomo to generate)
    :param file_name_prefix: bliss file prefix name
    :param file_name_z_fill: optional z fill for the file name index. If None then file index will not be 'z filled'
    :param create_tomo_config: if True create the 'tomo_config' group under instrument which contains
                                    metadata describing the acquisition (which dataset to read for rotation, translation ...)
    """

    def __init__(
        self,
        n_sample,
        n_sequence,
        n_scan_per_sequence,
        n_darks,
        n_flats,
        output_dir,
        with_nx_detector_attr=True,
        detector_name="pcolinux",
        acqui_type="basic",
        z_values=None,
        nb_loop=None,
        nb_tomo=None,
        nb_turns=None,
        with_rotation_motor_info=True,
        frame_data_type=numpy.uint16,
        file_name_prefix="sample",
        file_name_z_fill=None,
        create_tomo_config: bool = True,
        ebs_tomo_version: str | None = None,
        z_series_v_3_options=None,
        rotation_is_clockwise: bool = False,
    ):
        self._n_darks = n_darks
        self._n_flats = n_flats
        self._n_scan_per_sequence = n_scan_per_sequence
        self.__folder = output_dir
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        self.__proposal_file = os.path.join(self.__folder, "ihproposal_file.h5")
        if acqui_type not in ("pcotomo", "multitomo"):
            if nb_loop is not None or nb_tomo is not None:
                raise ValueError(
                    "nb_loop and nb_tomo are only handled by acqui_type: `multitomo`"
                )
        else:
            if not (
                nb_turns is not None or (nb_loop is not None and nb_tomo is not None)
            ):
                raise ValueError(
                    "nb_turns should be provided or nb_loop and nb_tomo must be provided"
                )

        # create sample
        self.__samples = []
        for sample_i in range(n_sample):
            if file_name_z_fill is None:
                dir_name = f"{file_name_prefix}_{sample_i}"
            else:
                dir_name = f"{file_name_prefix}_{str(sample_i).zfill(file_name_z_fill)}"
            sample_dir = os.path.join(self.path, dir_name)
            os.mkdir(sample_dir)
            sample_file = os.path.join(sample_dir, dir_name + ".h5")
            if acqui_type == "basic":
                acqui_tomo = _BlissBasicTomo(
                    sample_dir=sample_dir,
                    sample_file=sample_file,
                    n_sequence=n_sequence,
                    n_scan_per_sequence=n_scan_per_sequence,
                    n_darks=n_darks,
                    n_flats=n_flats,
                    with_nx_detector_attr=with_nx_detector_attr,
                    detector_name=detector_name,
                    with_rotation_motor_info=with_rotation_motor_info,
                    frame_data_type=frame_data_type,
                    create_tomo_config=create_tomo_config,
                    ebs_tomo_version=ebs_tomo_version,
                    rotation_is_clockwise=rotation_is_clockwise,
                )
            elif acqui_type in ("pcotomo", "multitomo"):
                acqui_tomo = _BlissMultiTomo(
                    sample_dir=sample_dir,
                    sample_file=sample_file,
                    n_sequence=n_sequence,
                    n_scan_per_sequence=n_scan_per_sequence,
                    n_darks=n_darks,
                    n_flats=n_flats,
                    with_nx_detector_attr=with_nx_detector_attr,
                    detector_name=detector_name,
                    with_rotation_motor_info=with_rotation_motor_info,
                    nb_loop=nb_loop,
                    nb_tomo=nb_tomo,
                    nb_turns=nb_turns,
                    frame_data_type=frame_data_type,
                    create_tomo_config=create_tomo_config,
                    ebs_tomo_version=ebs_tomo_version,
                    rotation_is_clockwise=rotation_is_clockwise,
                )
            elif acqui_type in ("z-series-v1", "z-series-v3"):
                if z_values is None:
                    raise ValueError("for z-series z_values should be provided")
                acqui_tomo = _BlissZseriesTomo(
                    sample_dir=sample_dir,
                    sample_file=sample_file,
                    n_sequence=n_sequence,
                    n_scan_per_sequence=n_scan_per_sequence,
                    n_darks=n_darks,
                    n_flats=n_flats,
                    with_nx_detector_attr=with_nx_detector_attr,
                    detector_name=detector_name,
                    z_values=z_values,
                    with_rotation_motor_info=with_rotation_motor_info,
                    frame_data_type=frame_data_type,
                    create_tomo_config=create_tomo_config,
                    ebs_tomo_version=ebs_tomo_version,
                    z_series_version=acqui_type.split("-")[-1],
                    z_series_v_3_options=z_series_v_3_options,
                    rotation_is_clockwise=rotation_is_clockwise,
                )
            else:
                raise NotImplementedError("")
            self.__samples.append(acqui_tomo)

    @property
    def samples(self):
        return self.__samples

    @property
    def proposal_file(self):
        # for now a simple file
        return self.__proposal_file

    @property
    def path(self):
        return self.__folder


class _BlissSample:
    """
    Simple mock of a bliss sample. For now we only create the hierarchy of
    files.
    """

    def __init__(
        self,
        sample_dir,
        sample_file,
        n_sequence,
        n_scan_per_sequence,
        n_darks,
        n_flats,
        detector_name,
        with_nx_detector_attr=True,
        with_rotation_motor_info=True,
        frame_data_type=numpy.uint16,
        create_tomo_config: bool = True,
        ebs_tomo_version: str | None = None,
        rotation_is_clockwise: bool = False,
    ):
        self._with_nx_detector_attr = with_nx_detector_attr
        self._sample_dir = sample_dir
        self._sample_file = sample_file
        self._n_sequence = n_sequence
        self._n_scan_per_seq = n_scan_per_sequence
        self._n_darks = n_darks
        self._n_flats = n_flats
        self._scan_folders = []
        self._index = 1
        self._detector_name = detector_name
        self._det_width = 64
        self._det_height = 64
        self._tomo_n = 10
        self._energy = 19.0
        self._sample_detector_distance = 100.0  # in mm
        self._source_sample_distance = 52000.0  # in mm
        self._sample_pixel_size = (0.0065, 0.0066)
        self._detector_pixel_size = (0.00022, 0.00022)
        self._with_rotation_motor_info = with_rotation_motor_info
        self._frame_data_type = frame_data_type
        self._create_tomo_config = create_tomo_config
        self._ebs_tomo_version = ebs_tomo_version
        assert isinstance(rotation_is_clockwise, bool)
        self._rotation_is_clockwise = rotation_is_clockwise
        for _ in range(n_sequence):
            self.add_sequence()

    @property
    def frame_data_type(self):
        return self._frame_data_type

    def get_next_free_index(self):
        idx = self._index
        self._index += 1
        return idx

    @property
    def current_scan_index(self) -> int:
        return self._index - 1

    def get_main_entry_title(self):
        raise NotImplementedError("Base class")

    @staticmethod
    def get_title(scan_type):
        if scan_type == "dark":
            return "dark images"
        elif scan_type == "flat":
            return "reference images 1"
        elif scan_type == "projection":
            return "projections 1 - 2000"
        else:
            raise ValueError("Not implemented")

    def create_entry_and_technique(self, seq_ini_index):
        # add sequence init information
        with HDF5File(self.sample_file, mode="a") as h5f:
            seq_node = h5f.require_group(str(seq_ini_index) + ".1")
            seq_node.attrs["NX_class"] = "NXentry"
            seq_node["title"] = self.get_main_entry_title()
            seq_node.require_group("instrument/positioners")
            # write energy
            seq_node["technique/scan/energy"] = self._energy
            seq_node["technique/scan/tomo_n"] = self._tomo_n * self._n_scan_per_seq
            seq_node["technique/scan/sample_detector_distance"] = (
                self._sample_detector_distance
            )
            seq_node["technique/scan/sample_detector_distance"].attrs["units"] = "mm"
            seq_node["technique/scan/source_sample_distance"] = (
                self._source_sample_distance
            )
            seq_node["technique/scan/source_sample_distance"].attrs["units"] = "mm"
            seq_node["technique/optic/sample_pixel_size"] = numpy.asarray(
                self._sample_pixel_size
            )
            seq_node["technique/detector/pixel_size"] = numpy.asarray(
                self._detector_pixel_size
            )
            seq_node["start_time"] = str(datetime.datetime.now())
            seq_node["end_time"] = str(
                datetime.datetime.now() + datetime.timedelta(minutes=10)
            )
            if self._create_tomo_config:
                self._add_tomo_config(seq_node)

    @staticmethod
    def get_next_group_name(seq_ini_index, scan_idx):
        return str(scan_idx) + ".1"

    def add_scan(
        self,
        scan_type,
        seq_ini_index,
        z_value,
        skip_title=False,
        nb_loop=None,
        nb_tomo=None,
        nb_turns=1,
    ):
        """
        :param nb_loop: number of loop in multitomo use case. Else must be 1
        :param nb_tomo: number of tomography done in multitomo 'per iteration' use case. Else must be 1
        """
        scan_idx = self.get_next_free_index()
        scan_name = str(scan_idx).zfill(4)
        scan_path = os.path.join(self.path, scan_name)
        self._scan_folders.append(_BlissScan(folder=scan_path, scan_type=scan_type))
        if nb_turns is not None:
            nb_nxtomo = nb_turns
            if nb_tomo is not None or nb_loop is not None:
                raise ValueError(
                    "nb_tomo and nb_loop should be provided or nb_turns. Not both"
                )
        elif nb_loop is not None and nb_tomo is not None:
            nb_nxtomo = nb_loop * nb_tomo
            if nb_turns is not None:
                raise ValueError(
                    "nb_tomo and nb_loop should be provided or nb_turns. Not both"
                )
        else:
            raise ValueError(
                "nb_tomo and nb_loop should be provided or nb_turns. None provided"
            )
        # register the scan information
        with HDF5File(self.sample_file, mode="a") as h5f:
            seq_node = h5f.require_group(str(scan_idx) + ".1")
            if "start_time" not in seq_node:
                seq_node["start_time"] = str(datetime.datetime.now())
            # write title
            title = self.get_title(scan_type=scan_type)
            if not skip_title:
                seq_node["title"] = title
            # write data
            data = (
                numpy.random.random(
                    self._det_height * self._det_width * self._tomo_n * nb_nxtomo
                )
                * 256
            )
            n_frames = self._tomo_n * nb_nxtomo
            data = data.reshape(n_frames, self._det_height, self._det_width)
            data = data.astype(self.frame_data_type)
            det_path_1 = "/".join(("instrument", self._detector_name))
            det_grp = seq_node.require_group(det_path_1)
            det_grp["data"] = data
            if self._with_nx_detector_attr:
                det_grp.attrs["NX_class"] = "NXdetector"
            acq_grp = det_grp.require_group("acq_parameters")
            acq_grp["acq_expo_time"] = 4
            det_path_2 = "/".join(("technique", "scan", self._detector_name))
            seq_node[det_path_2] = data
            seq_node.attrs["NX_class"] = "NXentry"

            # write rotation angle value and translations
            instrument_group = seq_node.require_group("instrument")
            positioners_grp = instrument_group.require_group("positioners")
            positioners_grp["hrsrot"] = numpy.linspace(
                start=0.0, stop=360, num=n_frames
            )
            positioners_grp["sx"] = numpy.array(numpy.random.random(size=n_frames))
            positioners_grp["sy"] = numpy.random.random(size=n_frames)
            positioners_grp["sz"] = numpy.asarray([z_value] * n_frames)
            positioners_grp["yrot"] = numpy.random.random(size=n_frames)
            if self._with_rotation_motor_info:
                scan_node = seq_node.require_group("technique/scan")
                scan_node["motor"] = ("rotation", "hrsrot", "srot")

            if self._ebs_tomo_version is not None:
                technique_group = seq_node.require_group("technique")
                technique_group.attrs["tomo_version"] = self._ebs_tomo_version

    def _add_tomo_config(self, group: h5py.Group):
        technique_group = group.require_group("technique")
        tomo_config_group = technique_group.require_group("tomoconfig")
        tomo_config_group["rotation"] = ["hrsrot"]
        tomo_config_group["detector"] = [
            self._detector_name,
        ]
        tomo_config_group["sample_x"] = ["sx"]
        tomo_config_group["sample_y"] = ["sy"]
        tomo_config_group["translation_z"] = ["sz"]
        tomo_config_group["translation_y"] = ["yrot"]
        tomo_config_group["rotation_is_clockwise"] = self._rotation_is_clockwise

    def add_sequence(self):
        """Add a sequence to the bliss file"""
        raise NotImplementedError("Base class")

    @property
    def path(self):
        return self._sample_dir

    @property
    def sample_directory(self):
        return self._sample_dir

    @property
    def sample_file(self):
        return self._sample_file

    def scans_folders(self):
        return self._scan_folders

    @property
    def n_darks(self):
        return self._n_darks

    @property
    def with_rotation_motor_info(self):
        return self._with_rotation_motor_info


class _BlissScan:
    """
    mock of a bliss scan
    """

    def __init__(self, folder, scan_type: str):
        assert scan_type in ("dark", "flat", "projection")
        self.__path = folder

    def path(self):
        return self.__path


class _BlissBasicTomo(_BlissSample):
    def get_main_entry_title(self):
        return "tomo:fullturn"

    def add_sequence(self):
        # reserve the index for the 'initialization' sequence. No scan folder
        # will be created for this one.
        seq_ini_index = self.get_next_free_index()

        self.create_entry_and_technique(seq_ini_index=seq_ini_index)

        if self.n_darks > 0:
            self.add_scan(scan_type="dark", seq_ini_index=seq_ini_index, z_value=1)

        if self._n_flats > 0:
            self.add_scan(scan_type="flat", seq_ini_index=seq_ini_index, z_value=1)

        for _ in range(self._n_scan_per_seq):
            self.add_scan(
                scan_type="projection", seq_ini_index=seq_ini_index, z_value=1
            )


class _BlissMultiTomo(_BlissSample):
    def __init__(
        self,
        sample_dir,
        sample_file,
        n_sequence,
        n_scan_per_sequence,
        n_darks,
        n_flats,
        detector_name,
        with_nx_detector_attr=True,
        with_rotation_motor_info=True,
        nb_loop=None,
        nb_tomo=None,
        nb_turns=1,
        frame_data_type=numpy.uint16,
        create_tomo_config: bool = True,
        ebs_tomo_version=None,
        rotation_is_clockwise: bool = False,
    ):
        self.nb_loop = nb_loop
        self.nb_tomo = nb_tomo
        self.nb_turns = nb_turns
        super().__init__(
            sample_dir,
            sample_file,
            n_sequence,
            n_scan_per_sequence,
            n_darks,
            n_flats,
            detector_name,
            with_nx_detector_attr,
            with_rotation_motor_info=with_rotation_motor_info,
            frame_data_type=frame_data_type,
            create_tomo_config=create_tomo_config,
            ebs_tomo_version=ebs_tomo_version,
            rotation_is_clockwise=rotation_is_clockwise,
        )
        if nb_loop is not None and nb_tomo is not None:
            if nb_turns is not None:
                raise ValueError(
                    "All of nb_loop, nb_tomo and nb_turns provided. Unable to deduce the pcotomo version"
                )
            multitomo_version = 1
        elif nb_turns is not None:
            multitomo_version = 2
        else:
            multitomo_version = None
        if multitomo_version is not None:
            # write Bliss version in attrs
            with HDF5File(self.sample_file, mode="a") as h5f:
                if "creator_version" not in h5f.attrs:
                    if multitomo_version == 1:
                        h5f.attrs["creator_version"] = "1.2.3"
                    if multitomo_version == 2:
                        h5f.attrs["creator_version"] = "1.10.0"

    def get_main_entry_title(self):
        return "tomo:multitomo"

    def add_sequence(self):
        # reserve the index for the 'initialization' sequence. No scan folder
        # will be created for this one.
        seq_ini_index = self.get_next_free_index()

        self.create_entry_and_technique(seq_ini_index=seq_ini_index)

        # start dark
        if self.n_darks > 0:
            self.add_scan(scan_type="dark", seq_ini_index=seq_ini_index, z_value=1)

        # start flat
        if self._n_flats > 0:
            self.add_scan(scan_type="flat", seq_ini_index=seq_ini_index, z_value=1)

        for _ in range(self._n_scan_per_seq):
            self.add_scan(
                scan_type="projection",
                seq_ini_index=seq_ini_index,
                z_value=1,
                nb_loop=self.nb_loop,
                nb_tomo=self.nb_tomo,
                nb_turns=self.nb_turns,
            )

        # end flat
        if self._n_flats > 0:
            self.add_scan(scan_type="flat", seq_ini_index=seq_ini_index, z_value=1)

    def add_scan(
        self,
        scan_type,
        seq_ini_index,
        z_value,
        skip_title=False,
        nb_loop=None,
        nb_tomo=None,
        nb_turns=1,
    ):
        super().add_scan(
            scan_type, seq_ini_index, z_value, skip_title, nb_loop, nb_tomo, nb_turns
        )
        if scan_type == "projection":
            # register multi-tomo specific informations (only for projections)
            with HDF5File(self.sample_file, mode="a") as h5f:
                seq_node = h5f.require_group(str(self._index - 1) + ".1")

                scan_grp = seq_node.require_group("technique/proj")
                if nb_loop is not None and "nb_loop" not in scan_grp:
                    scan_grp["nb_loop"] = nb_loop
                if nb_tomo is not None and "nb_tomo" not in scan_grp:
                    scan_grp["nb_tomo"] = nb_tomo
                if nb_turns is not None and "nb_turns" not in scan_grp:
                    scan_grp["nb_turns"] = nb_turns
                if "tomo_n" not in scan_grp:
                    scan_grp["tomo_n"] = self._tomo_n


class _BlissZseriesTomo(_BlissSample):
    def __init__(
        self,
        sample_dir,
        sample_file,
        n_sequence,
        n_scan_per_sequence,
        n_darks,
        n_flats,
        detector_name,
        z_values,
        z_series_version: str,
        with_nx_detector_attr=True,
        with_rotation_motor_info=True,
        frame_data_type=numpy.uint16,
        create_tomo_config: bool = True,
        ebs_tomo_version: str = None,
        z_series_v_3_options=None,
        rotation_is_clockwise: bool = False,
    ):
        assert z_series_version in ("v1", "v3")
        self.z_series_version = z_series_version
        self._z_values = z_values
        self._z_series_v_3_options = z_series_v_3_options
        super().__init__(
            sample_dir=sample_dir,
            sample_file=sample_file,
            n_sequence=n_sequence,
            n_scan_per_sequence=n_scan_per_sequence,
            n_darks=n_darks,
            n_flats=n_flats,
            detector_name=detector_name,
            with_nx_detector_attr=with_nx_detector_attr,
            with_rotation_motor_info=with_rotation_motor_info,
            frame_data_type=frame_data_type,
            create_tomo_config=create_tomo_config,
            ebs_tomo_version=ebs_tomo_version,
            rotation_is_clockwise=rotation_is_clockwise,
        )

    def get_main_entry_title(self):
        return "tomo:zseries"

    def create_dark_at_start(self) -> bool:
        if self.z_series_version == "v1":
            return True
        else:
            return self._z_series_v_3_options["dark_at_start"]

    def create_flat_at_start(self) -> bool:
        if self.z_series_version == "v1":
            return True
        else:
            return self._z_series_v_3_options["flat_at_start"]

    def create_dark_at_end(self) -> bool:
        if self.z_series_version == "v1":
            return True
        else:
            return self._z_series_v_3_options["dark_at_end"]

    def create_flat_at_end(self) -> bool:
        if self.z_series_version == "v1":
            return True
        else:
            return self._z_series_v_3_options["flat_at_end"]

    def create_intermediary_flat(self) -> bool:
        return self.z_series_version == "v1"

    def create_intermediary_dark(self) -> bool:
        return self.z_series_version == "v1"

    def add_sequence(self):
        # reserve the index for the 'initialization' sequence. No scan folder
        # will be created for this one.

        if self.z_series_version == "v1":
            seq_ini_index = self.get_next_free_index()
            self.create_entry_and_technique(seq_ini_index=seq_ini_index)

        for z_value in self._z_values:
            if self.z_series_version == "v3":
                seq_ini_index = self.get_next_free_index()
                self.create_entry_and_technique(seq_ini_index=seq_ini_index)

            if z_value == self._z_values[0]:
                create_dark = self.create_dark_at_start()
                create_flat = self.create_flat_at_start()
            elif z_value == self._z_values[-1]:
                create_dark = self.create_dark_at_end()
                create_flat = self.create_flat_at_end()
            else:
                create_dark = self.create_intermediary_dark()
                create_flat = self.create_intermediary_flat()

            if create_dark and self.n_darks > 0:
                self.add_scan(
                    scan_type="dark", seq_ini_index=seq_ini_index, z_value=z_value
                )

            if create_flat and self._n_flats > 0:
                self.add_scan(
                    scan_type="flat", seq_ini_index=seq_ini_index, z_value=z_value
                )

            for _ in range(self._n_scan_per_seq):
                self.add_scan(
                    scan_type="projection", seq_ini_index=seq_ini_index, z_value=z_value
                )

    def create_entry_and_technique(self, seq_ini_index):
        super().create_entry_and_technique(seq_ini_index=seq_ini_index)
        # add sequence init information
        if self.z_series_version == "v3":
            with HDF5File(self.sample_file, mode="a") as h5f:
                seq_node = h5f.require_group(str(seq_ini_index) + ".1")
                seq_node.attrs["NX_class"] = "NXentry"
                # write scab flags
                seq_node["technique/scan_flags/dark_images_at_start"] = (
                    self._z_series_v_3_options["dark_at_start"]
                )
                seq_node["technique/scan_flags/dark_images_at_end"] = (
                    self._z_series_v_3_options["dark_at_end"]
                )
                seq_node["technique/scan_flags/ref_images_at_start"] = (
                    self._z_series_v_3_options["flat_at_start"]
                )
                seq_node["technique/scan_flags/ref_images_at_end"] = (
                    self._z_series_v_3_options["flat_at_end"]
                )

    def add_scan(
        self,
        scan_type,
        seq_ini_index,
        z_value,
        skip_title=False,
        nb_loop=None,
        nb_tomo=None,
        nb_turns=1,
    ):
        super().add_scan(
            scan_type, seq_ini_index, z_value, skip_title, nb_loop, nb_tomo, nb_turns
        )
        with HDF5File(self.sample_file, mode="a") as h5f:
            seq_node = h5f.require_group(str(self.current_scan_index) + ".1")
            seq_node["sample/name"] = "mysample_0000"



# coding: utf-8

"""
module to convert from edf to (nexus tomo compliant) .nx
"""

from __future__ import annotations

import logging
import os
from collections import namedtuple
import fabio
import h5py
import numpy
import pint

from tqdm import tqdm

from silx.io.dictdump import dicttoh5
from silx.utils.deprecation import deprecated

from tomoscan.esrf.scan.edfscan import EDFTomoScan
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf.scan.utils import get_parameters_frm_par_or_info
from tomoscan.io import HDF5File

from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.nxobject.nxdetector import FieldOfView, ImageKey
from nxtomo.nxobject.nxsource import SourceType

from nxtomomill import utils
from nxtomomill.converter.version import LATEST_VERSION
from nxtomomill.converter.version import version as converter_version
from nxtomomill.models.utils import PathType
from nxtomomill.models.edf2nx import EDF2nxModel
from nxtomomill.utils.nexus import create_nx_data_group, link_nxbeam_to_root
from nxtomomill.settings import Tomo

__all__ = [
    "EDFFileKeys",
    "DEFAULT_EDF_KEYS",
    "edf_to_nx",
    "from_edf_to_nx",
    "post_processing_check",
    "get_byte_order",
]

EDF_MOTOR_POS = Tomo.EDF.MOTOR_POS
EDF_MOTOR_MNE = Tomo.EDF.MOTOR_MNE
EDF_REFS_NAMES = Tomo.EDF.REFS_NAMES
EDF_TO_IGNORE = Tomo.EDF.TO_IGNORE
EDF_ROT_ANGLE = Tomo.EDF.ROT_ANGLE
EDF_DARK_NAMES = Tomo.EDF.DARK_NAMES
EDF_X_TRANS = Tomo.EDF.X_TRANS
EDF_Y_TRANS = Tomo.EDF.Y_TRANS
EDF_Z_TRANS = Tomo.EDF.Z_TRANS
EDF_MACHINE_CURRENT = Tomo.EDF.MACHINE_CURRENT

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)


EDFFileKeys = namedtuple(
    "EDFFileKeys",
    [
        "motor_pos_keys",
        "motor_mne_keys",
        "rotation_angle_keys",
        "x_translation_keys",
        "y_translation_keys",
        "z_translation_keys",
        "to_ignore",
        "dark_names",
        "flat_names",
        "machine_current_keys",
    ],
)

DEFAULT_EDF_KEYS = EDFFileKeys(
    EDF_MOTOR_POS,
    EDF_MOTOR_MNE,
    EDF_ROT_ANGLE,
    EDF_X_TRANS,
    EDF_Y_TRANS,
    EDF_Z_TRANS,
    EDF_TO_IGNORE,
    EDF_DARK_NAMES,
    EDF_REFS_NAMES,
    EDF_MACHINE_CURRENT,
)

_ESTIMATED_PERCENTAGE_TIME_TO_READ_METADATA = 20
# as at the moment we cannot provide a good estimation of how much it takes to read metadata let's estimate it from experience


@deprecated(replacement="from_edf_to_nx", since_version="0.9.0")
def edf_to_nx(
    scan: EDFTomoScan,
    output_file: str,
    file_extension: str,
    file_keys: EDFFileKeys = DEFAULT_EDF_KEYS,
    progress=None,
    sample_name: str | None = None,
    title: str | None = None,
    instrument_name: str | None = None,
    source_name: str | None = None,
    source_type: SourceType | None = None,
) -> tuple:
    """
    Convert an edf file to a nexus file.
    For now duplicate data.

    :param scan:
    :param output_file:
    :param file_extension:
    :param file_keys:
    :param progress:
    :param sample_name: name of the sample
    :param title: dataset title
    :param instrument_name: name of the instrument used
    :param source_name: name of the source (most likely ESRF)
    :param source_type: type of the source (most likely "Synchrotron X-ray Source")
    :return: (nexus_file, entry)
    """
    if not isinstance(scan, EDFTomoScan):
        raise TypeError("scan is expected to be an instance of EDFTomoScan")

    config = EDF2nxModel(
        input_folder=scan.path,
        dataset_basename=scan.dataset_basename,
        output_file=output_file,
        file_extension=file_extension,
        sample_name=sample_name,
        title=title,
        instrument_name=instrument_name,
        source_name=source_name,
        source_type=source_type,
        # handle file_keys
        motor_position_keys=file_keys.motor_pos_keys,
        motor_mne_keys=file_keys.motor_mne_keys,
        rotation_angle_keys=file_keys.rotation_angle_keys,
        x_translation_keys=file_keys.x_translation_keys,
        y_translation_keys=file_keys.y_translation_keys,
        z_translation_keys=file_keys.z_translation_keys,
        machine_current_keys=file_keys.machine_current_keys,
        patterns_to_ignores=file_keys.to_ignore,
        dark_names_prefix=file_keys.dark_names,
        flat_names_prefix=file_keys.flat_names,
    )

    return from_edf_to_nx(config, progress=progress)


def from_edf_to_nx(configuration: EDF2nxModel, progress: tqdm | None = None) -> tuple:
    """
    Convert an edf file to a nexus file.
    For now duplicate data.

    :param configuration: configuration to use to process the data
    :param progress: if provided then will be updated with conversion progress
    :return: (nexus_file, entry)
    """
    if configuration.input_folder is None:
        raise ValueError("input_folder should be provided")
    if not os.path.isdir(configuration.input_folder):
        raise OSError(f"{configuration.input_folder} is not a valid folder path")

    if configuration.output_file is None:
        raise ValueError("output_file should be provided")

    # if we don't duplicate data then we can't delete sources EDF files
    if not configuration.duplicate_data and configuration.delete_edf_source_files:
        raise ValueError(
            "You asked for avoiding data duplication and to delete edf source files. "
            "Those two options are not compatible. Avoiding data duplication will "
            "lead to create HDF5Virtual dataset pointing to the edf source files"
        )

    fileout_h5 = utils.get_file_name(
        file_name=configuration.output_file,
        extension=configuration.file_extension,
        check=True,
    )

    with HDF5File(fileout_h5, "w") as h5d:
        return __process(
            configuration=configuration,
            output_grp=h5d,
            fileout_h5=fileout_h5,
            progress=progress,
        )


def post_processing_check(configuration: EDF2nxModel):
    """
    check that conversion made contains the same information as a folder and optionnaly delete edf files

    """
    if configuration.input_folder is None:
        raise ValueError("input_folder should be provided")
    if not os.path.isdir(configuration.input_folder):
        raise OSError(f"{configuration.input_folder} is not a valid folder path")

    if configuration.output_file is None:
        raise ValueError("output_file should be provided")
    if (
        configuration.delete_edf_source_files is True
        and len(configuration.output_checks) == 0
    ):
        raise ValueError(
            "requested to remove edf files without righting an NXtomo and not doing any check on an existing NXtomo. This is a non sense."
        )

    # if we don't duplicate data then we can't delete sources EDF files
    if not configuration.duplicate_data and configuration.delete_edf_source_files:
        raise ValueError(
            "You asked for avoiding data duplication and to delete edf source files. "
            "Those two options are not compatible. Avoiding data duplication will "
            "lead to create HDF5Virtual dataset pointing to the edf source files"
        )

    fileout_h5 = utils.get_file_name(
        file_name=configuration.output_file,
        extension=configuration.file_extension,
        check=True,
    )

    return __process(
        configuration=configuration,
        fileout_h5=fileout_h5,
        output_grp=None,
        progress=None,
    )


def __process(
    configuration: EDF2nxModel,
    output_grp: h5py.Group | None,
    fileout_h5: str,
    progress: tqdm | None,
):
    """
    This function is used to insure processing is exactly the same:
     * if we want to convert edf to .nx
     * if we want to check a conversion from edf to edf (and optionally remove edf afterwards)

    Not very nice but the simpler for making evolve some legacy code...
    """

    if progress is not None:
        progress.set_postfix_str(
            "preprocessing - retrieve all metadata (can take a few seconds - cannot display real advancement)"
        )
        progress.total = 100
        progress.n = _ESTIMATED_PERCENTAGE_TIME_TO_READ_METADATA
        progress.refresh()

    if configuration.dataset_info_file is not None:
        if not os.path.isfile(configuration.dataset_info_file):
            raise ValueError(f"{configuration.dataset_info_file} is not a file")
        else:
            scan_info = get_parameters_frm_par_or_info(configuration.dataset_info_file)
    else:
        scan_info = None
    scan = EDFTomoScan(
        scan=configuration.input_folder,
        dataset_basename=configuration.dataset_basename,
        scan_info=scan_info,
        # TODO: add n frames ?
    )

    _logger.info(f"Output file will be {fileout_h5}")
    if scan.dim_2 is None or scan.dim_1 is None:
        if scan_info is not None:
            raise ValueError("Please provide scan 'dim_1' and 'dim_2' with scan info")
        else:
            info_file = os.path.join(
                configuration.input_folder, f"{configuration.dataset_basename}.info"
            )
            raise ValueError(
                f"Unable to deduce scan dimension from scan metadata (used file for metadata: {info_file})"
            )

    default_current = scan.retrieve_information(
        scan=scan.path,
        dataset_basename=scan.dataset_basename,
        ref_file=None,
        key="SrCurrent",
        key_aliases=["SRCUR", "machineCurrentStart"],
        type_=float,
        scan_info=scan.scan_info,
    )
    if default_current is None:
        _logger.warning("Unable to find machine current default current. Take 0.0")
        default_current = 0.0

    output_data_path = "entry"

    # with today design we can simply remove the files of the url. We don't expect those files
    # to contain other type of data
    edf_source_files = dict()
    # for each file register a tuple (existing_frame, set(url)) so at the end we can make sure all the frame have been used before
    # removing the file
    detector_data_urls = []
    # store the url in the order they are used and if it has been modified or not (for output checks).

    def update_edf_source_files(url, n_frames):
        if configuration.delete_edf_source_files:
            # edf_source files is only used for deleting edf files
            edf_file_path = url.file_path()
            file_source_info = edf_source_files.get(edf_file_path, (n_frames, set()))
            urls_used = file_source_info[1]
            urls_used.add(url.path())
            edf_source_files[edf_file_path] = (
                file_source_info[0],
                urls_used,
            )

    DARK_ACCUM_FACT = True

    metadata = []
    proj_urls = scan.get_proj_urls(
        scan=scan.path, dataset_basename=scan.dataset_basename
    )

    for dark_to_find in configuration.dark_names_prefix:
        dk_urls = scan.get_darks_url(scan_path=scan.path, prefix=dark_to_find)
        if len(dk_urls) > 0:
            if dark_to_find == "dark":
                DARK_ACCUM_FACT = False
            break

    _edf_to_ignore = list(configuration.patterns_to_ignores)

    for refs_to_find in configuration.flat_names_prefix:
        if refs_to_find == "ref":
            _edf_to_ignore.append("HST")
        elif "HST" in _edf_to_ignore:
            _edf_to_ignore.remove("HST")

        refs_urls = scan.get_flats_url(
            scan_path=scan.path,
            prefix=refs_to_find,
            ignore=_edf_to_ignore,
            dataset_basename=scan.dataset_basename,
        )
        if len(refs_urls) > 0:
            break

    n_frames = len(proj_urls) + len(refs_urls) + len(dk_urls)
    n_darks = len(dk_urls)

    (
        frame_type,
        rot_angle_index,
        x_trans_index,
        y_trans_index,
        z_trans_index,
        srcur_index,
    ) = _getExtraInfo(scan=scan, configuration=configuration)

    if rot_angle_index == -1 and configuration.force_angle_calculation is False:
        _logger.warning(
            f"Unable to find one of the defined key for rotation in header ({configuration.rotation_angle_keys}). Will force angle calculation"
        )
        configuration.force_angle_calculation = True

    if not output_grp:
        # in the case we don't intend to write the NXtomo but only check it
        output_grp = None
        ext_datasets_grp = None
        dark_dataset = None
        data_dataset = None
        keys_dataset = None
        keys_control_dataset = None
        x_dataset = y_dataset = z_dataset = None
        rotation_dataset = None
        machine_current_dataset = None
    else:
        if configuration.duplicate_data is True:
            data_dataset = output_grp.create_dataset(
                "/entry/instrument/detector/data",
                shape=(n_frames, scan.dim_2, scan.dim_1),
                dtype=frame_type,
            )
            ext_datasets_grp = None
            dark_dataset = None
        else:
            _logger.warning(
                "No data duplication requested. Will fail if your dataset contains compressed data"
            )
            # if necessary create external datasets groups
            # note: for now we are force to create one dataset per frame because
            # the byte order can be modified from one frame to the other.
            ext_datasets_grp = output_grp.create_group("external_datasets")
            data_dataset = None
            # we must duplicate darks not matter what because exposure time is not handled
            # at nabu side for the moment
            dark_dataset = ext_datasets_grp.create_dataset(
                "darks",
                shape=(n_darks, scan.dim_2, scan.dim_1),
                dtype=frame_type,
            )

        keys_dataset = output_grp.create_dataset(
            "/entry/instrument/detector/image_key", shape=(n_frames,), dtype=numpy.int32
        )

        keys_control_dataset = output_grp.create_dataset(
            "/entry/instrument/detector/image_key_control",
            shape=(n_frames,),
            dtype=numpy.int32,
        )

        title = configuration.title
        if title is None:
            title = os.path.basename(scan.path)
        output_grp["/entry/title"] = title

        sample_name = configuration.sample_name
        if configuration.sample_name is None:
            # try to deduce sample name from scan path.
            try:
                sample_name = os.path.abspath(scan.path).split(os.sep)[-3:]
                sample_name = os.sep.join(sample_name)
            except Exception:
                sample_name = "unknow"
        output_grp["/entry/sample/name"] = sample_name
        if configuration.instrument_name is not None:
            instrument_grp = output_grp["/entry"].require_group("instrument")
            instrument_grp["name"] = configuration.instrument_name

        if configuration.source_name is not None:
            source_grp = output_grp["/entry/instrument"].require_group("source")
            source_grp["name"] = configuration.source_name
        if configuration.source_type is not None:
            source_grp = output_grp["/entry/instrument"].require_group("source")
            source_grp["type"] = configuration.source_type.value
        if configuration.source_probe is not None:
            source_grp = output_grp["/entry/instrument"].require_group("source")
            source_grp["probe"] = configuration.source_probe.value

        if scan.scan_range is None or scan.tomo_n is None:
            raise ValueError(
                f"Cannot find scan_range ({scan.scan_range}) and / or tomo_n ({scan.tomo_n}). Is the .info file here and / or valid ?"
            )

        distance = scan.retrieve_information(
            scan=os.path.abspath(scan.path),
            dataset_basename=scan.dataset_basename,
            ref_file=None,
            key="Distance",
            type_=float,
            key_aliases=["distance"],
            scan_info=scan.scan_info,
        )
        if distance is not None:
            output_grp["/entry/instrument/detector/distance"] = (
                (distance * configuration.sample_detector_distance_unit)
                .to(_ureg.meter)
                .magnitude
            )
            output_grp["/entry/instrument/detector/distance"].attrs["units"] = str(
                (_ureg.meter)
            )

        pixel_size = scan.retrieve_information(
            scan=os.path.abspath(scan.path),
            dataset_basename=scan.dataset_basename,
            ref_file=None,
            key="PixelSize",
            type_=float,
            key_aliases=["pixelSize"],
            scan_info=scan.scan_info,
        )
        output_grp["/entry/instrument/detector/x_pixel_size"] = (
            (pixel_size * configuration.pixel_size_unit).to(_ureg.meter).magnitude
        )
        output_grp["/entry/instrument/detector/x_pixel_size"].attrs["units"] = str(
            str(_ureg.meter)
        )
        output_grp["/entry/instrument/detector/y_pixel_size"] = (
            (pixel_size * configuration.pixel_size_unit).to(_ureg.meter).magnitude
        )
        output_grp["/entry/instrument/detector/y_pixel_size"].attrs["units"] = str(
            str(_ureg.meter)
        )

        energy = scan.retrieve_information(
            scan=os.path.abspath(scan.path),
            dataset_basename=scan.dataset_basename,
            ref_file=None,
            key="Energy",
            type_=float,
            key_aliases=["energy"],
            scan_info=scan.scan_info,
        )
        if energy is not None:
            energy = energy * configuration.energy_unit
            output_grp["/entry/instrument/beam/incident_energy"] = energy.to(
                _ureg.keV
            ).magnitude
            output_grp["/entry/instrument/beam/incident_energy"].attrs["units"] = str(
                _ureg.keV
            )

        # rotations values
        rotation_dataset = output_grp.create_dataset(
            "/entry/sample/rotation_angle", shape=(n_frames,), dtype=numpy.float32
        )
        output_grp["/entry/sample/rotation_angle"].attrs["units"] = str(_ureg.degree)

        # machine current
        nexus_paths = get_nexus_paths(LATEST_VERSION)
        machine_current_path = "/".join(["entry", nexus_paths.ELECTRIC_CURRENT_PATH])
        machine_current_dataset = output_grp.create_dataset(
            machine_current_path,
            shape=(n_frames,),
            dtype=numpy.float32,
        )
        output_grp[machine_current_path].attrs["units"] = str(_ureg.ampere)

        # provision for centering motors
        # warning: x_dataset, y_dataset and z_dataset are from the esrf reference.
        # from NXtomo reference this is z_translation, x_translation and y_translation
        x_dataset = output_grp.create_dataset(
            "/entry/sample/z_translation", shape=(n_frames,), dtype=numpy.float32
        )
        output_grp["/entry/sample/z_translation"].attrs["units"] = str(_ureg.meter)

        y_dataset = output_grp.create_dataset(
            "/entry/sample/x_translation", shape=(n_frames,), dtype=numpy.float32
        )
        output_grp["/entry/sample/x_translation"].attrs["units"] = str(_ureg.meter)

        z_dataset = output_grp.create_dataset(
            "/entry/sample/y_translation", shape=(n_frames,), dtype=numpy.float32
        )
        output_grp["/entry/sample/y_translation"].attrs["units"] = str(_ureg.meter)

    #  --------->  and now fill all datasets!

    nf = 0

    external_datasets = []  # collect all urls created
    progress_v = 0
    if progress is not None:
        progress.set_postfix_str("write dark")

    def ignore(file_name):
        for forbid in _edf_to_ignore:
            if forbid in file_name:
                return True
        return False

    # darks

    # dark in acumulation mode?
    norm_dark = 1.0
    if scan.dark_n > 0 and DARK_ACCUM_FACT is True:
        norm_dark = len(dk_urls) / scan.dark_n
    dk_indexes = sorted(dk_urls.keys())

    for dk_index in dk_indexes:
        dk_url = dk_urls[dk_index]
        if ignore(os.path.basename(dk_url.file_path())):
            _logger.info("ignore " + dk_url.file_path())
            continue
        data, header, external_dataset, n_frames_in_file = _read_url(
            url=dk_url,
            i_frame=dk_index,
            h5group_to_dump=None,  # never create external dataset for dark
            frame_prefix="dark",
            configuration=configuration,
        )
        assert header is not None
        metadata.append(header)

        update_edf_source_files(url=dk_url, n_frames=n_frames_in_file)
        detector_data_urls.append((dk_url, True))
        if output_grp:
            if configuration.duplicate_data:
                data_dataset[nf, :, :] = data * norm_dark
            else:
                dark_dataset[nf, :, :] = data * norm_dark
            keys_dataset[nf] = ImageKey.DARK_FIELD.value
            keys_control_dataset[nf] = ImageKey.DARK_FIELD.value

            motor_pos_key = _get_valid_key(header, configuration.motor_position_keys)
            if motor_pos_key:
                str_mot_val = header[motor_pos_key].split(" ")
                if rot_angle_index == -1 or configuration.force_angle_calculation:
                    rotation_dataset[nf] = 0.0
                else:
                    rotation_dataset[nf] = float(str_mot_val[rot_angle_index])
                if x_trans_index == -1:
                    x_dataset[nf] = 0.0
                else:
                    x_dataset[nf] = (
                        (
                            float(str_mot_val[x_trans_index])
                            * configuration.x_translation_unit
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )
                if y_trans_index == -1:
                    y_dataset[nf] = 0.0
                else:
                    y_dataset[nf] = (
                        (
                            float(str_mot_val[y_trans_index])
                            * configuration.y_translation_unit
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )
                if z_trans_index == -1:
                    z_dataset[nf] = 0.0
                else:
                    z_dataset[nf] = (
                        (
                            float(str_mot_val[z_trans_index])
                            * configuration.z_translation_unit
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )

            if srcur_index == -1:
                machine_current_dataset[nf] = (
                    (default_current * configuration.machine_current_unit)
                    .to(_ureg.ampere)
                    .magnitude
                )
            else:
                try:
                    str_counter_val = header.get("counter_pos", "").split(" ")
                    machine_current_dataset[nf] = (
                        (
                            float(str_counter_val[srcur_index])
                            * configuration.machine_current_unit
                        )
                        .to(_ureg.ampere)
                        .magnitude
                    )
                except IndexError:
                    machine_current_dataset[nf] = (
                        (default_current * configuration.machine_current_unit)
                        .to(_ureg.ampere)
                        .magnitude
                    )

            nf += 1
            if progress is not None:
                progress_v += 1 / (n_frames - 1)
                progress.n = (
                    _ESTIMATED_PERCENTAGE_TIME_TO_READ_METADATA
                    + progress_v * (100 - _ESTIMATED_PERCENTAGE_TIME_TO_READ_METADATA)
                )
                progress.refresh()

    ref_indexes = sorted(refs_urls.keys())

    ref_projs = []
    for irf in ref_indexes:
        pjnum = int(irf)
        if pjnum not in ref_projs:
            ref_projs.append(pjnum)

    # refs
    def store_refs(
        refIndexes,
        projnum,
        refUrls,
        nF,
        dataDataset,
        keysDataset,
        keysCDataset,
        xDataset,
        yDataset,
        zDataset,
        rotationDataset,
        raix,
        xtix,
        ytix,
        ztix,
    ):
        nfr = nF
        for ref_index in refIndexes:
            int_rf = int(ref_index)
            if int_rf == projnum:
                refUrl = refUrls[ref_index]
                if ignore(os.path.basename(refUrl.file_path())):
                    _logger.info("ignore " + refUrl.file_path())
                    continue
                data, header, external_data, n_frames_in_file = _read_url(
                    url=refUrl,
                    i_frame=ref_index,
                    h5group_to_dump=ext_datasets_grp,
                    frame_prefix="flat",
                    configuration=configuration,
                )
                update_edf_source_files(url=refUrl, n_frames=n_frames_in_file)
                if external_data is not None:
                    external_datasets.append(external_data)
                metadata.append(header)
                detector_data_urls.append((refUrl, False))

                if output_grp is None:
                    continue

                if configuration.duplicate_data and dataDataset:
                    dataDataset[nfr, :, :] = data
                keysDataset[nfr] = ImageKey.FLAT_FIELD.value
                keysCDataset[nfr] = ImageKey.FLAT_FIELD.value
                motor_pos_key = _get_valid_key(
                    header, configuration.motor_position_keys
                )

                if motor_pos_key in header:
                    str_mot_val = header[motor_pos_key].split(" ")
                    if raix == -1 or configuration.force_angle_calculation:
                        rotationDataset[nfr] = 0.0
                    else:
                        rotationDataset[nfr] = float(str_mot_val[raix])
                    if xtix == -1:
                        xDataset[nfr] = 0.0
                    else:
                        xDataset[nfr] = (
                            (
                                float(str_mot_val[xtix])
                                * configuration.x_translation_unit
                            )
                            .to(_ureg.meter)
                            .magnitude
                        )
                    if ytix == -1:
                        yDataset[nfr] = 0.0
                    else:
                        yDataset[nfr] = (
                            (
                                float(str_mot_val[ytix])
                                * configuration.y_translation_unit
                            )
                            .to(_ureg.meter)
                            .magnitude
                        )
                    if ztix == -1:
                        zDataset[nfr] = 0.0
                    else:
                        zDataset[nfr] = (
                            (
                                float(str_mot_val[ztix])
                                * configuration.z_translation_unit
                            )
                            .to(_ureg.meter)
                            .magnitude
                        )
                if srcur_index == -1:
                    machine_current_dataset[nfr] = (
                        (default_current * configuration.machine_current_unit)
                        .to(_ureg.ampere)
                        .magnitude
                    )
                else:
                    str_counter_val = header.get("counter_pos", "").split(" ")
                    try:
                        machine_current_dataset[nfr] = (
                            (
                                float(str_counter_val[srcur_index])
                                * configuration.machine_current_unit
                            )
                            .to(_ureg.ampere)
                            .magnitude
                        )
                    except IndexError:
                        machine_current_dataset[nfr] = (
                            (default_current * configuration.machine_current_unit)
                            .to(_ureg.ampere)
                            .magnitude
                        )

                nfr += 1

        return nfr

    # projections
    proj_indexes = sorted(proj_urls.keys())
    if progress is not None:
        progress.set_postfix_str("write projections and flats")
    nproj = 0
    iref_pj = 0

    if configuration.force_angle_calculation:
        if configuration.angle_calculation_rev_neg_scan_range and scan.scan_range < 0:
            proj_angles = numpy.linspace(
                0,
                scan.scan_range,
                scan.tomo_n,
                endpoint=configuration.angle_calculation_endpoint,
            )
        else:
            proj_angles = numpy.linspace(
                min(0, scan.scan_range),
                max(0, scan.scan_range),
                scan.tomo_n,
                endpoint=configuration.angle_calculation_endpoint,
            )
            revert_angles_in_nx = (
                configuration.angle_calculation_rev_neg_scan_range
                and (scan.scan_range < 0)
            )
            if revert_angles_in_nx:
                proj_angles = proj_angles[::-1]

    alignment_indices = []
    for i_proj, proj_index in enumerate(proj_indexes):
        proj_url = proj_urls[proj_index]
        if ignore(os.path.basename(proj_url.file_path())):
            _logger.info("ignore " + proj_url.file_path())
            continue

        # store refs if the ref serial number is = projection number
        if iref_pj < len(ref_projs) and ref_projs[iref_pj] == nproj:
            nf = store_refs(
                ref_indexes,
                ref_projs[iref_pj],
                refs_urls,
                nf,
                data_dataset,
                keys_dataset,
                keys_control_dataset,
                x_dataset,
                y_dataset,
                z_dataset,
                rotation_dataset,
                rot_angle_index,
                x_trans_index,
                y_trans_index,
                z_trans_index,
            )
            iref_pj += 1
        data, header, external_data, n_frames_in_file = _read_url(
            proj_url,
            i_frame=proj_index,
            h5group_to_dump=ext_datasets_grp,
            frame_prefix="proj",
            configuration=configuration,
        )
        update_edf_source_files(url=proj_url, n_frames=n_frames_in_file)
        detector_data_urls.append((proj_url, False))

        if output_grp:
            if external_data is not None:
                external_datasets.append(external_data)
            metadata.append(header)

            if configuration.duplicate_data:
                data_dataset[nf, :, :] = data
            keys_dataset[nf] = ImageKey.PROJECTION.value
            keys_control_dataset[nf] = ImageKey.PROJECTION.value
            if nproj >= scan.tomo_n:
                keys_control_dataset[nf] = ImageKey.ALIGNMENT.value

            motor_pos_key = _get_valid_key(header, configuration.motor_position_keys)

            if configuration.force_angle_calculation:
                if i_proj < len(proj_angles):
                    rotation_dataset[nf] = proj_angles[i_proj]
                else:
                    # case of return / control projection
                    rotation_dataset[nf] = 0.0

            if motor_pos_key in header:
                str_mot_val = header[motor_pos_key].split(" ")

                # continuous scan - rot angle is unknown. Compute it
                if not configuration.force_angle_calculation:
                    rotation_dataset[nf] = float(str_mot_val[rot_angle_index])

                if x_trans_index == -1:
                    x_dataset[nf] = 0.0
                else:
                    x_dataset[nf] = (
                        (
                            float(str_mot_val[x_trans_index])
                            * configuration.x_translation_unit
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )
                if y_trans_index == -1:
                    y_dataset[nf] = 0.0
                else:
                    y_dataset[nf] = (
                        (
                            float(str_mot_val[y_trans_index])
                            * configuration.y_translation_unit
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )
                if z_trans_index == -1:
                    z_dataset[nf] = 0.0
                else:
                    z_dataset[nf] = (
                        (
                            float(str_mot_val[z_trans_index])
                            * configuration.z_translation_unit
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )
            if srcur_index == -1:
                machine_current_dataset[nf] = (
                    (default_current * configuration.machine_current_unit)
                    .to(_ureg.ampere)
                    .magnitude
                )
            else:
                try:
                    str_counter_val = header.get("counter_pos", "").split(" ")
                    machine_current_dataset[nf] = (
                        (
                            float(str_counter_val[srcur_index])
                            * configuration.machine_current_unit
                        )
                        .to(_ureg.ampere)
                        .magnitude
                    )
                except IndexError:
                    machine_current_dataset[nf] = (
                        (default_current * configuration.machine_current_unit)
                        .to(_ureg.ampere)
                        .magnitude
                    )
        nf += 1
        nproj += 1

        if progress is not None:
            progress_v += 1 / (n_frames - 1)
            progress.n = _ESTIMATED_PERCENTAGE_TIME_TO_READ_METADATA + progress_v * (
                100 - _ESTIMATED_PERCENTAGE_TIME_TO_READ_METADATA
            )
            progress.refresh()

    # we need to update alignement angles values. I wanted to avoid to redo all the previous existing processing.
    n_alignment_angles = len(alignment_indices)
    if n_alignment_angles == 3:
        alignments_angles = numpy.linspace(
            scan.scan_range,
            0,
            n_alignment_angles,
            endpoint=(n_alignment_angles % 2) == 0,
        )
        for index, angle in zip(alignment_indices, alignments_angles):
            rotation_dataset[index] = angle

    # store last flat if any remaining in the list
    if iref_pj < len(ref_projs) and output_grp:
        nf = store_refs(
            ref_indexes,
            ref_projs[iref_pj],
            refs_urls,
            nf,
            data_dataset,
            keys_dataset,
            keys_control_dataset,
            x_dataset,
            y_dataset,
            z_dataset,
            rotation_dataset,
            rot_angle_index,
            x_trans_index,
            y_trans_index,
            z_trans_index,
        )

    if output_grp:
        # if we avoided data duplication: create the virtual dataset on top of the external datasets
        if not configuration.duplicate_data:
            virtual_layout = h5py.VirtualLayout(
                shape=(n_frames, scan.dim_2, scan.dim_1),
                dtype=frame_type,
            )
            virtual_layout[0:n_darks] = h5py.VirtualSource(dark_dataset)
            for i_ext, external_dataset in enumerate(external_datasets):
                assert isinstance(external_dataset, h5py.Dataset)
                virtual_layout[i_ext + n_darks] = h5py.VirtualSource(external_dataset)
            output_grp.create_virtual_dataset(
                "/entry/instrument/detector/data",
                virtual_layout,
            )

        # we can add some more NeXus look and feel
        output_grp["/entry"].attrs["NX_class"] = "NXentry"
        output_grp["/entry"].attrs["definition"] = "NXtomo"
        output_grp["/entry"].attrs["version"] = converter_version()
        output_grp["/entry/instrument"].attrs["NX_class"] = "NXinstrument"
        output_grp["/entry/instrument/detector"].attrs["NX_class"] = "NXdetector"
        output_grp["/entry/instrument/detector/data"].attrs["interpretation"] = "image"
        if configuration.field_of_view is not None:
            field_of_view = configuration.field_of_view
        elif abs(scan.scan_range) == 180:
            field_of_view = "Full"
        elif abs(scan.scan_range) == 360:
            field_of_view = "Half"

        if field_of_view is not None:
            field_of_view = FieldOfView(field_of_view)
            output_grp["/entry/instrument/detector/field_of_view"] = field_of_view.value

        output_grp["/entry/sample"].attrs["NX_class"] = "NXsample"
        output_grp["/entry/definition"] = "NXtomo"
        source_grp = output_grp["/entry/instrument"].get("source", None)
        if source_grp is not None and "NX_class" not in source_grp.attrs:
            source_grp.attrs["NX_class"] = "NXsource"

        output_grp.flush()

        for i_meta, meta in enumerate(metadata):
            # save metadata
            dicttoh5(
                meta,
                fileout_h5,
                h5path=f"/entry/instrument/positioners/{i_meta}",
                update_mode="replace",
                mode="a",
            )

        try:
            create_nx_data_group(
                file_path=fileout_h5,
                entry_path=output_data_path,
                axis_scale=["linear", "linear"],
            )
        except Exception as e:
            _logger.error(f"Fail to create NXdata group. Reason is {e}")

        # create beam group at root for compatibility
        try:
            link_nxbeam_to_root(file_path=fileout_h5, entry_path=output_data_path)
        except Exception as e:
            _logger.error(f"Fail to link nx beam. Error is {e}")
        if progress is not None:
            print("\nconversion finished\n")

    if output_grp is not None:
        # if the output group is provided then close it so we can safely read it back
        # it should be done outside this function but this is legacy code that won't evolve
        output_grp.close()
    # do some check on the conversion
    issues_discovered = []
    for output_check in configuration.output_checks:
        print("\nstart output checks\n")
        from nxtomomill.converter.edf import checks as _checks_utils

        output_check = _checks_utils.OUPUT_CHECK(output_check)
        if output_check is _checks_utils.OUPUT_CHECK.COMPARE_VOLUME:
            issues = _checks_utils.compare_volumes(
                edf_volume_as_urls=detector_data_urls,
                hdf5_scan=NXtomoScan(fileout_h5, output_data_path),
            )
        else:
            raise ValueError(f"{output_check} is not handled")
        issues_discovered.extend(issues)
    if len(issues_discovered) > 0:
        raise ValueError("Seems the conversion failed. Detected issues are {errs}")
    elif len(configuration.output_checks) > 0:
        print("\noutput checks done, no issues detected\n")

    if configuration.delete_edf_source_files:
        print("start edf source files removal")
        assert (
            configuration.duplicate_data
        ), "if data is not duplicated this make no sense to remove the edf files"
        for file_path, (n_frames, used_urls) in edf_source_files.items():
            if n_frames > len(used_urls):
                _logger.warning(
                    f"Will not delete {file_path}. Only {len(used_urls)} used on the {n_frames} contained"
                )
            else:
                try:
                    os.remove(file_path)
                except OSError as e:
                    _logger.error(f"Issue when removing {file_path}. Error is {e}")

    return fileout_h5, output_data_path


def _get_valid_key(header: dict, keys: tuple) -> str | None:
    """Return the first existing key in header"""
    for key in keys:
        if key in header:
            return key
    else:
        return None


def _get_valid_key_index(motors: list, keys: tuple) -> int | None:
    for key in keys:
        if key in motors:
            return motors.index(key)
    else:
        return -1


def _read_url(
    url,
    h5group_to_dump: h5py.Group,
    i_frame: int,
    frame_prefix: str,
    configuration: EDF2nxModel,
) -> tuple:
    """
    if h5group_to_dump is provided then it will create a dataset with external data to
    the EDF file under the name `frame_{i}`
    This way we will be able to create a virtual
    """
    data_slice = url.data_slice()
    if data_slice is None:
        data_slice = (0,)
    if data_slice is None or len(data_slice) != 1:
        raise ValueError(f"Fabio slice expect a single frame but {data_slice} found")
    index = data_slice[0]
    if not isinstance(index, int):
        raise ValueError(f"Fabio slice expect a single integer but {data_slice} found")

    try:
        fabio_file = fabio.open(url.file_path())
    except Exception:
        _logger.debug(
            f"Error while opening {url.file_path()} with fabio", exc_info=True
        )
        raise IOError(
            f"Error while opening {url.path()} with fabio (use debug for more information)"
        )

    try:
        n_frames_in_file = fabio_file.nframes
        if n_frames_in_file == 1:
            if index != 0:
                raise ValueError(
                    f"Only a single frame available. Slice {index} out of range"
                )
            it = fabio.edfimage.EdfImage.lazy_iterator(url.file_path())
            frame = next(it)
        else:
            frame = fabio_file.getframe(index)
        data = frame.data
        header = frame.header
        frame_start_offset = frame.start
        blobsize = frame.blobsize
        byte_order = get_byte_order(frame.header)
    except Exception as e:
        _logger.error(e)
        data = None
        header = None
        frame_start_offset = None
        blobsize = None
        byte_order = None
    finally:
        fabio_file.close()
        fabio_file = None

    if data is not None and byte_order is None:
        raise ValueError("Unable to get ByteOrder for file_path")

    if h5group_to_dump is not None and data is not None:
        frame_start = frame_start_offset + blobsize - (data.size * data.dtype.itemsize)
        frame_end = frame_start + blobsize
        dataset_name = f"{frame_prefix}_{i_frame}"
        data_type = numpy.dtype(byte_order + data.dtype.char)
        if configuration.external_link_type is PathType.ABSOLUTE:
            file_path = os.path.abspath(url.file_path())
        elif configuration.external_link_type is PathType.RELATIVE:
            file_path = os.path.abspath(url.file_path())
            file_path = os.path.relpath(
                os.path.abspath(file_path),
                os.path.abspath(os.path.dirname(configuration.output_file)),
            )
            file_path = "./" + file_path
        else:
            raise NotImplementedError
        external_dataset = h5group_to_dump.create_dataset(
            name=dataset_name,
            shape=data.shape,
            dtype=data_type,
            external=[
                (file_path, frame_start, frame_end),
            ],
        )
    else:
        external_dataset = None
    return data, header, external_dataset, n_frames_in_file


def _getExtraInfo(scan, configuration):
    assert isinstance(scan, EDFTomoScan)
    projections_urls = scan.projections
    if len(projections_urls) == 0:
        raise ValueError(
            f"No projections found in {scan.path} with dataset basename: {configuration.dataset_basename if configuration.dataset_basename is not None else 'Default'} and dataset info file: {configuration.dataset_info_file if configuration.dataset_info_file is not None else 'Default'}. "
        )
    indexes = sorted(projections_urls.keys())
    first_proj_file = projections_urls[indexes[0]]
    fid = fabio.open(first_proj_file.file_path())

    rotangle_index = -1
    xtrans_index = -1
    ytrans_index = -1
    ztrans_index = -1
    srcur_index = -1
    frame_type = None

    try:
        if hasattr(fid, "header"):
            hd = fid.header
        else:
            hd = fid.getHeader()
        motor_mne_key = _get_valid_key(hd, configuration.motor_mne_keys)
        motors = hd.get(motor_mne_key, "").split(" ")
        counters = hd.get("counter_mne", "").split(" ")
        rotangle_index = _get_valid_key_index(motors, configuration.rotation_angle_keys)
        xtrans_index = _get_valid_key_index(motors, configuration.x_translation_keys)
        ytrans_index = _get_valid_key_index(motors, configuration.y_translation_keys)
        ztrans_index = _get_valid_key_index(motors, configuration.z_translation_keys)
        srcur_index = _get_valid_key_index(counters, configuration.machine_current_keys)

        if hasattr(fid, "bytecode"):
            frame_type = fid.bytecode
        else:
            frame_type = fid.getByteCode()
    finally:
        fid.close()
        fid = None

    return (
        frame_type,
        rotangle_index,
        xtrans_index,
        ytrans_index,
        ztrans_index,
        srcur_index,
    )


def get_byte_order(header):
    """
    byte_order (as a str compatible with numpy data types)
    """
    byte_order = header.get("ByteOrder", None)
    if byte_order is None:
        pass
    elif byte_order.lower() == "highbytefirst":
        byte_order = ">"
    elif byte_order.lower() == "lowbytefirst":
        byte_order = "<"
    return byte_order



# coding: utf-8

"""
Base class for tomography acquisition (defined by Bliss)
"""

from __future__ import annotations

import os
import pint

import h5py

from nxtomo.application.nxtomo import NXtomo
from nxtomomill.utils.h5pyutils import from_data_url_to_virtual_source
from nxtomomill.utils.hdf5 import EntryReader, get_dataset_unit
from nxtomomill.utils.utils import embed_url, get_file_name  # noqa F401
from nxtomomill.utils._os import abspath

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
import logging
from collections import OrderedDict

import numpy
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset

from nxtomomill.io.config import TomoHDF5Config

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)

__all__ = ["BaseAcquisition", "get_dataset_name_from_motor"]


def _ask_for_file_removal(file_path):
    res = input(f"Overwrite {file_path} ? (y/N)").lower()
    return res == "y"


class BaseAcquisition:
    """
    Util class to group several hdf5 group together and to write the data
    Nexus / NXtomo compliant
    """

    _ENERGY_PATH = "technique/scan/energy"

    _DATASET_NAME_PATH = ("technique/scan/name",)

    _GRP_SIZE_PATH = ("technique/scan/nb_scans",)

    _SAMPLE_NAME_PATH = ("sample/name",)

    TITLE_PATHS = (
        "technique/scan/sequence",
        "title",
    )

    _INSTRUMENT_NAME_PATH = (
        "technique/saving/beamline",
        "instrument/title",
    )

    _FOV_PATH = "technique/scan/field_of_view"

    _NB_LOOP_PATH = ("technique/scan/nb_loop", "technique/proj/nb_loop")

    _NB_TOMO_PATH = ("technique/scan/nb_tomo", "technique/proj/nb_tomo")

    _NB_TURNS_PATH = (
        "technique/proj/nb_turns",
        "technique/scan/nb_turns",
    )

    _TOMO_N_PATH = (
        "technique/proj/tomo_n",
        "technique/scan/tomo_n",
        "technique/proj/proj_n",
        "technique/scan/proj_n",
    )

    _START_TIME_PATH = ("start_time",)

    _END_TIME_PATH = ("end_time",)

    _TECHNIQUE_MOTOR_PATHS = ("technique/scan/motor", "technique/proj/motor")

    _DETECTOR_ROI = ("technique/detector/roi",)

    _SOURCE_NAME = ("instrument/machine/name",)

    _SOURCE_TYPE = ("instrument/machine/type",)

    _FRAME_FLIP_PATHS = (
        "technique/detector/{detector_name}/flipping",
        "technique/detector/flipping",
    )

    _PROPAGATION_DISTANCE_PATHS = (
        "technique/scan/effective_propagation_distance",
        "technique/scan/propagation_distance",
    )

    def __init__(
        self,
        root_url: DataUrl | None,
        configuration: TomoHDF5Config,
        detector_sel_callback,
        start_index: int,
    ):
        self._root_url = root_url
        self._detector_sel_callback = detector_sel_callback
        self._registered_entries = OrderedDict()
        self._copy_frames = OrderedDict()
        # key is the entry, value his type
        self._entries_o_path = dict()
        # key is the entry, value his the entry path from the original file.
        # this value is different from `.name`. Don't know if this is a bug ?
        """user can have defined already some parameter values as energy.
        The idea is to avoid asking him if """
        self._configuration = configuration
        self._start_index = start_index

        # set of properties when an acquisition needs to copy dark / flat (used in the case of z-series version 3)
        self._dark_at_start = None
        self._dark_at_end = None
        self._flat_at_start = None
        self._flat_at_end = None

    @property
    def configuration(self):
        return self._configuration

    @property
    def start_index(self) -> int:
        return self._start_index

    def write_as_nxtomo(
        self,
        shift_entry: int,
        input_file_path: str,
        request_input: bool,
        divide_into_sub_files: bool,
        input_callback=None,
    ) -> tuple:
        """
        This function will dump the acquisition to disk as an NXtomo

        :param shift_entry: index of the entry to start saving new nxtomos.
        :param input_file_path: output file path
        :param request_input: if True the conversion can ask user some missing metadata
        :param divide_into_sub_files: if True then create one file per NXtomo
        :param input_callback: function to call for users to provide missing metadata
        """
        nx_tomos = self.to_NXtomos(
            request_input=request_input,
            input_callback=input_callback,
            check_tomo_n=True,
        )

        # preprocessing to define output file name
        possible_extensions = (".hdf5", ".h5", ".nx", ".nexus")
        output_file_basename = os.path.basename(self.configuration.output_file)
        file_extension_ = None
        for possible_extension in possible_extensions:
            if output_file_basename.endswith(possible_extension):
                output_file_basename.rstrip(possible_extension)
                file_extension_ = possible_extension

        def get_file_name_and_entry(index, divide_sub_files):
            entry = "entry" + str(index).zfill(4)

            if self.configuration.single_file or not divide_sub_files:
                en_output_file = self.configuration.output_file
            else:
                ext = file_extension_ or self.configuration.file_extension
                file_name = (
                    os.path.splitext(output_file_basename)[0]
                    + "_"
                    + str(index).zfill(4)
                    + ext
                )
                en_output_file = os.path.join(
                    os.path.dirname(self.configuration.output_file), file_name
                )

                if os.path.exists(en_output_file):
                    if self.configuration.overwrite is True:
                        _logger.warning(en_output_file + " will be removed")
                        _logger.info("remove " + en_output_file)
                        os.remove(en_output_file)
                    elif _ask_for_file_removal(en_output_file) is False:
                        raise OSError(f"unable to overwrite {en_output_file}, exit")
                    else:
                        os.remove(en_output_file)
            return en_output_file, entry

        result = []
        for i_nx_tomo, nx_tomo in enumerate(nx_tomos):
            output_file, data_path = get_file_name_and_entry(
                (shift_entry + i_nx_tomo), divide_sub_files=divide_into_sub_files
            )
            output_file = abspath(os.path.relpath(output_file, os.getcwd()))
            output_file = abspath(output_file)

            # For multi-tomo for example the data path is modified in order to handle with the splitting
            # that does not fit at 100 % with the current API. for now there is
            # no convenient way to handle this in a better way
            assert isinstance(nx_tomo, NXtomo)
            # embed data if requested
            if len(nx_tomo.instrument.detector.data) == 0:
                _logger.warning(
                    f"No frame found for NXtomo number {i_nx_tomo}. Won't write any output for it"
                )
                continue
            vs_0 = nx_tomo.instrument.detector.data[0]
            if nx_tomo.detector_data_is_defined_by_url():
                new_urls = []
                for url in nx_tomo.instrument.detector.data:
                    if self._copy_frames[url.path()]:
                        created_url = embed_url(
                            url=url,
                            output_file=output_file,
                        )
                        new_urls.append(created_url)
                    else:
                        new_urls.append(url)
                nx_tomo.instrument.detector.data = new_urls
            elif nx_tomo.detector_data_is_defined_by_virtual_source():
                new_vs = []
                for vs in nx_tomo.instrument.detector.data:
                    assert isinstance(vs, h5py.VirtualSource)
                    assert isinstance(vs_0, h5py.VirtualSource)
                    url = DataUrl(file_path=vs.path, data_path=vs.name, scheme="silx")
                    if (
                        url.path() in self._copy_frames
                        and self._copy_frames[url.path()]
                    ):
                        new_url = embed_url(url, output_file=output_file)
                        n_vs, _, _ = from_data_url_to_virtual_source(new_url)
                        new_vs.append(n_vs)
                    else:
                        new_vs.append(vs)
                nx_tomo.instrument.detector.data = new_vs

            # provide some extra data like origin of the input_file
            if input_file_path is not None:
                nx_tomo.bliss_original_files = (abspath(input_file_path),)
            # save data
            nx_tomo.save(file_path=output_file, data_path=data_path)
            # check rotation angle
            if nx_tomo.sample.rotation_angle is not None:
                unique_angles = numpy.unique(
                    nx_tomo.sample.rotation_angle.to(_ureg.degree).magnitude
                )
                if len(unique_angles) == 1:
                    _logger.warning(
                        f"NXtomo {data_path}@{output_file} seems to have a single value ({unique_angles}) for rotation angle. Seems it fails to find correct path to the rotation angle dataset"
                    )

            result.append((output_file, data_path))
        return tuple(result)

    def to_NXtomos(self, request_input, input_callback, check_tomo_n=True) -> tuple:
        raise NotImplementedError("Base class")

    @property
    def raise_error_if_issue(self):
        """
        Should we raise an error if we encounter or an issue or should we
        just log an error message
        """
        return self.configuration.raises_error

    @property
    def root_url(self):
        return self._root_url

    def get_expected_nx_tomo(self):
        """
        Return the expected number of nxtomo created for this acquisition.
        This is required to get consistent entry and file name. At lest for automation
        """
        raise NotImplementedError("Base class")

    def read_entry(self):
        return EntryReader(self._root_url)

    def is_different_sequence(self, entry):
        """
        Can we have several entries 1.1, 1.2, 1.3... to consider.
        """
        raise ValueError("Base class")

    def is_part_of_same_series(self, other: BaseAcquisition) -> bool:
        if not isinstance(other, BaseAcquisition):
            raise TypeError("Can only compar two z-series")
        if self.root_url is None or other.root_url is None:
            return False

        def read_sample_node(url: DataUrl) -> str | None:
            with EntryReader(url=url) as entry:
                sample_node = self._get_sample_node(entry)
                if sample_node is None:
                    return None
                sample_name = sample_node.get("name", None)
                if sample_name is not None:
                    return h5py_read_dataset(sample_name)
                return None

        self_sample_name = read_sample_node(self.root_url)
        other_sample_name = read_sample_node(other.root_url)
        if self_sample_name is None or other_sample_name is None:
            return False

        def filter_scan_index(sample_name: str) -> str:
            # for z_series the sample name is post pone with _{sample_name} that we need to filter
            parts = sample_name.split("_")
            if len(parts) > 1 and numpy.char.isnumeric(parts[-1]):
                return "_".join(parts[:-1])
            return sample_name

        return filter_scan_index(self_sample_name) == filter_scan_index(
            other_sample_name
        )

    @staticmethod
    def _get_node_values_for_frame_array(
        node: h5py.Group,
        n_frame: int | None,
        keys: list | tuple,
        info_retrieve: str,
        expected_unit: pint.Unit,
    ) -> tuple[numpy.ndarray | None, str]:
        assert isinstance(expected_unit, pint.Unit), "Invalid expected unit"

        def get_values():
            # this is a two step process: first step we parse all the
            # the keys until we found one with the expected length
            # if first iteration fails then we return the first existing key
            for respect_length in (True, False):
                for possible_key in keys:
                    if possible_key in node and isinstance(
                        node[possible_key], h5py.Dataset
                    ):
                        values = h5py_read_dataset(node[possible_key])
                        unit = get_dataset_unit(
                            dataset=node[possible_key],
                            default=expected_unit,
                            from_dataset=f"{node[possible_key].name} (reading {info_retrieve})",
                        )
                        # skip values containing '*DIS*'
                        if isinstance(values, str) and values == "*DIS*":
                            continue

                        if n_frame is not None and respect_length is True:
                            if numpy.isscalar(values):
                                length = 1
                            else:
                                length = len(values)
                            if length in (n_frame, n_frame + 1):
                                return values, unit
                        else:
                            return values, unit
            return None, None

        values, unit = get_values()
        if values is None:
            raise ValueError(
                f"Unable to retrieve {info_retrieve} for {node.name}. Was looking for {keys} datasets"
            )
        elif n_frame is None:
            return values, unit
        elif numpy.isscalar(values):
            return numpy.array([values] * n_frame), unit
        elif len(values) == n_frame:
            return values.tolist(), unit
        elif len(values) == (n_frame + 1):
            # for now we can have one extra position for rotation,
            # translation_x...
            # because saved after the last projection. It is recording the
            # motor position. For example in this case: 1 is the motor movement
            # (saved) and 2 is the acquisition
            #
            #  1     2    1    2     1
            #      -----     -----
            # -----     -----     -----
            #
            return values[:-1].tolist(), unit
        elif len(values) > n_frame:
            _logger.warning(
                f"Incoherent number of values found for {info_retrieve}. Can come from an acquisition canceled. Else please investigate."
            )
            # in this case only get the values which have a frame
            return values[0:n_frame], unit
        elif len(values) < n_frame:
            _logger.warning(
                f"Incoherent number of values found for {info_retrieve}. Can come from an acquisition canceled. Else please investigate."
            )
            # in this case append 0 to existing values. Maybe -1 would be better ?
            return list(values) + [0] * (n_frame - values), unit
        elif len(values) == 1:
            return numpy.array([values[0]] * n_frame), unit
        else:
            raise ValueError("incoherent number of angle position vs number of frame")

    def register_step(self, url: DataUrl, entry_type, copy_frames) -> None:
        """
        Add a bliss entry to the acquisition
        :param url:
        :param entry_type:
        """
        raise NotImplementedError("Base class")

    @staticmethod
    def _get_instrument_node(entry_node: h5py.Group) -> h5py.Group:
        if not isinstance(entry_node, h5py.Group):
            raise TypeError("entry_node: h5py.group expected")
        return entry_node["instrument"]

    @staticmethod
    def _get_positioners_node(entry_node: h5py.Group) -> h5py.Group | None:
        if not isinstance(entry_node, h5py.Group):
            raise TypeError("entry_node is expected to be a h5py.Group")
        parent_node = BaseAcquisition._get_instrument_node(entry_node)
        if "positioners" in parent_node:
            return parent_node["positioners"]
        else:
            return None

    @staticmethod
    def _get_measurement_node(entry_node: h5py.Group) -> h5py.Group | None:
        if not isinstance(entry_node, h5py.Group):
            raise TypeError("entry_node is expected to be a h5py.Group")
        if "measurement" in entry_node:
            return entry_node["measurement"]
        else:
            return None

    @staticmethod
    def _get_machine_node(entry_node: h5py.Group) -> h5py.Group | None:
        if not isinstance(entry_node, h5py.Group):
            raise TypeError("entry_node is expected to be a h5py.Group")
        if "instrument/machine" in entry_node:
            return entry_node["instrument/machine"]
        else:
            return None

    @staticmethod
    def _get_sample_node(entry_node: h5py.Group) -> h5py.Group | None:
        if not isinstance(entry_node, h5py.Group):
            raise TypeError("entry_node is expected to be a h5py.Group")
        if "sample" in entry_node:
            return entry_node["sample"]
        else:
            return None

    @staticmethod
    def _get_scan_flags(entry_node: h5py.Group) -> h5py.Group | None:
        if not isinstance(entry_node, h5py.Group):
            raise TypeError("entry_node is expected to be a h5py.Group")
        if "technique/scan_flags" in entry_node:
            return entry_node["technique/scan_flags"]
        else:
            return None

    def _read_rotation_motor_name(self) -> str | None:
        """read rotation motor from root_url/technique/scan/motor

        :return: name of the motor used for rotation. None if cannot find
        """
        if self._root_url is None:
            _logger.warning("no root url. Unable to read rotation motor")
            return None
        else:
            with EntryReader(self._root_url) as entry:
                for motor_path in self._TECHNIQUE_MOTOR_PATHS:
                    if motor_path in entry:
                        try:
                            rotation_motor = get_dataset_name_from_motor(
                                motors=h5py_read_dataset(
                                    numpy.asarray(entry[motor_path])
                                ),
                                motor_name="rotation",
                            )
                        except Exception as e:
                            _logger.error(e)
                        else:
                            return rotation_motor
                    else:
                        _logger.warning(
                            f"{motor_path} unable to find rotation motor from {self._root_url}"
                        )
                return None

    def _get_machine_current(self, root_node) -> None | pint.Quantity:
        """retrieve electric current provide a time stamp for each of them"""
        if root_node is None:
            _logger.warning("no root url. Unable to read electric current")
            return None
        else:
            grps = [
                root_node,
            ]

            measurement_node = self._get_measurement_node(root_node)
            if measurement_node is not None:
                grps.append(measurement_node)
            machine_node = self._get_machine_node(root_node)
            if machine_node is not None:
                grps.append(machine_node)

            for grp in grps:
                try:
                    mach_current, unit = self._get_node_values_for_frame_array(
                        node=grp,
                        keys=self.configuration.machine_current_keys,
                        info_retrieve="machine current",
                        expected_unit=_ureg.mA,
                        n_frame=None,
                    )
                except (ValueError, KeyError):
                    pass
                else:
                    # handle case where mach_current is a scalar. Cast it to list before return
                    if numpy.isscalar(mach_current):
                        mach_current = [
                            mach_current,
                        ]

                    return mach_current * unit
            else:
                _logger.warning(
                    f"Unable to retrieve machine current for {root_node.name}"
                )

            return None

    def _get_rotation_angle(self, root_node: h5py.Group, n_frame: int) -> pint.Quantity:
        """
        return the list of rotation angle for each frame.
        If unable to find it will either:
        * raise an error (if 'raise_error_if_issue' is True)
        * return a zeros for all required angles (kwnow from n_frame)
        """
        if not isinstance(root_node, h5py.Group):
            raise TypeError("root_node is expected to be a h5py.Group")

        for grp in (
            self._get_positioners_node(root_node),
            root_node,
            self._get_measurement_node(root_node),
        ):
            try:
                angles, unit = self._get_node_values_for_frame_array(
                    node=grp,
                    n_frame=n_frame,
                    keys=self.configuration.rotation_angle_keys,
                    info_retrieve="rotation angle",
                    expected_unit=_ureg.degree,
                )
            except (ValueError, KeyError):
                pass
            else:
                return angles * unit

        mess = f"Unable to find rotation angle for {root_node.name}"
        if self.raise_error_if_issue:
            raise ValueError(mess)
        else:
            mess += "default value will be set. (0)"
            _logger.warning(mess)
            return ([0] * n_frame) * _ureg.degree

    def _get_sample_x(self, root_node, n_frame) -> pint.Quantity:
        """return the list of translation for each frame"""
        for grp in self._get_positioners_node(root_node), root_node:
            try:
                x_tr, unit = self._get_node_values_for_frame_array(
                    node=grp,
                    n_frame=n_frame,
                    keys=self.configuration.sample_x_keys,
                    info_retrieve="sample x translation",
                    expected_unit=_ureg.mm,
                )
                x_tr = numpy.asarray(x_tr)
            except (ValueError, KeyError):
                pass
            else:
                return x_tr * unit

        mess = f"Unable to find sample x for {root_node.name}"
        if self.raise_error_if_issue:
            raise ValueError(mess)
        else:
            mess += "default value will be set. (0)"
            _logger.warning(mess)
            return ([0] * n_frame) * _ureg.meter

    def _get_sample_y(self, root_node, n_frame) -> pint.Quantity:
        """return the list of translation for each frame"""
        for grp in self._get_positioners_node(root_node), root_node:
            try:
                y_tr, unit = self._get_node_values_for_frame_array(
                    node=grp,
                    n_frame=n_frame,
                    keys=self.configuration.sample_y_keys,
                    info_retrieve="sample y translation",
                    expected_unit=_ureg.mm,
                )
                y_tr = numpy.asarray(y_tr)
            except (ValueError, KeyError):
                pass
            else:
                return y_tr * unit

        mess = f"Unable to find sample y for {root_node.name}"
        if self.raise_error_if_issue:
            raise ValueError(mess)
        else:
            mess += "default value will be set. (0)"
            _logger.warning(mess)
            return ([0] * n_frame) * _ureg.meter

    def _get_translation_y(self, root_node, n_frame) -> pint.quantity:
        """return the list of translation for each frame"""
        for grp in self._get_positioners_node(root_node), root_node:
            try:
                y_tr, unit = self._get_node_values_for_frame_array(
                    node=grp,
                    n_frame=n_frame,
                    keys=self.configuration.translation_y_keys,
                    info_retrieve="y base translation",
                    expected_unit=_ureg.mm,
                )
                y_tr = numpy.asarray(y_tr)
            except (ValueError, KeyError):
                pass
            else:
                return y_tr * unit

        mess = f"Unable to find translation y for {root_node.name}"
        if self.raise_error_if_issue:
            raise ValueError(mess)
        else:
            mess += "default value will be set. (0)"
            _logger.warning(mess)
            return ([0] * n_frame) * _ureg.meter

    @staticmethod
    def get_translation_z_frm(
        root_node, n_frame: int, configuration: TomoHDF5Config
    ) -> pint.Quantity:
        for grp in BaseAcquisition._get_positioners_node(root_node), root_node:
            try:
                z_tr, unit = BaseAcquisition._get_node_values_for_frame_array(
                    node=grp,
                    n_frame=n_frame,
                    keys=configuration.translation_z_keys,
                    info_retrieve="z translation",
                    expected_unit=_ureg.mm,
                )
                z_tr = numpy.asarray(z_tr)
            except (ValueError, KeyError):
                pass
            else:
                return z_tr * unit

        mess = f"Unable to find translation z on node {root_node.name}"
        if configuration.raises_error:
            raise ValueError(mess)
        else:
            mess += "default value will be set. (0)"
            _logger.warning(mess)
            return ([0] * n_frame) * _ureg.meter

    def _get_translation_z(self, root_node, n_frame) -> pint.Quantity:
        """return the list of translation z for each frame"""
        return self.get_translation_z_frm(
            root_node=root_node,
            n_frame=n_frame,
            configuration=self.configuration,
        )

    def _get_expo_time(self, root_node, n_frame, detector_node) -> pint.Quantity:
        """return expo time for each frame"""
        for grp in (detector_node.get("acq_parameters", None), root_node):
            if grp is None:
                continue
            try:
                expo, unit = self._get_node_values_for_frame_array(
                    node=grp,
                    n_frame=n_frame,
                    keys=self.configuration.exposure_time_keys,
                    info_retrieve="exposure time",
                    expected_unit=_ureg.second,
                )
            except (ValueError, KeyError):
                pass
            else:
                return expo * unit

        mess = f"Unable to find frame exposure time on entry {self.root_url.path()}"
        if self.raise_error_if_issue:
            raise ValueError(mess)
        else:
            mess += "default value will be set. (0)"
            _logger.warning(mess)
            return 0 * _ureg.second

    def get_axis_scale_types(self):
        """
        Return axis display for the detector data to be used by silx view
        """
        return ["linear", "linear"]

    def __str__(self):
        if self.root_url is None:
            return "NXTomo"
        else:
            return self.root_url.path()

    def get_detector_roi(self):
        if self._root_url is None:
            _logger.warning("no root url. Unable to read detector roi")
            return None
        else:
            with EntryReader(self._root_url) as entry:
                for roi_path in self._DETECTOR_ROI:
                    if roi_path in entry:
                        try:
                            roi = h5py_read_dataset(numpy.asarray(entry[roi_path]))
                        except Exception as e:
                            _logger.error(e)
                        else:
                            return roi
                    else:
                        _logger.warning(
                            f"{roi_path} unable to find detector roi from {self._root_url}"
                        )
                return None


def get_dataset_name_from_motor(motors, motor_name):
    motors = numpy.asarray(motors)
    indexes = numpy.where(motors == motor_name)[0]
    if len(indexes) == 0:
        return None
    elif len(indexes) == 1:
        index = indexes[0]
        index_dataset_id = index + 1
        if index_dataset_id < len(motors):
            return motors[index_dataset_id]
        else:
            raise ValueError(
                f"{motor_name} found but unable to find dataset name from {motors}"
            )
    else:
        raise ValueError(f"More than one instance of {motor_name} as been found.")
